import typing
from collections.abc import Generator, Iterable, Sequence
from contextlib import contextmanager
from functools import cached_property, wraps
from itertools import repeat
from typing import Literal, TypedDict

import numpy as np
import numpy.typing as npt

from napari.layers.shapes._mesh import Mesh
from napari.layers.shapes._shapes_constants import ShapeType, shape_classes
from napari.layers.shapes._shapes_models import Line, Path, Shape
from napari.layers.shapes._shapes_utils import triangles_intersect_box
from napari.layers.shapes.shape_types import (
    CoordinateArray,
    CoordinateDtype,
    IndexArray,
    IndexDtype,
    ShapeColor,
    ShapeColorArray,
    ShapeColorDtype,
    TriangleArray,
    TriangleDtype,
    ZOrderArray,
    ZOrderDtype,
)
from napari.utils.geometry import (
    inside_triangles,
    intersect_line_with_triangles,
    line_in_triangles_3d,
)
from napari.utils.translations import trans


class MeshArrayDict(TypedDict):
    """Mesh array dict used for adding multiple shapes at once.

    This type is helper for functions used when adding multiple shapes at once.
    To ensure names and types of arrays are consistent, we use a TypedDict.

    Fields
    ------
    z_index : ZOrderArray
        Information about z-ndex of shapes.
    vertices : CoordinateArray
        array of shape vertices
    mesh_vertices : CoordinateArray
        array of face triangulation vertices.
    mesh_vertices_centers : CoordinateArray
        array of face triangulation vertices centers.
        It is required for update of edges triangulation.
    mesh_vertices_offsets : CoordinateArray
        array of face triangulation vertices offsets.
        It is required for update of edges triangulation.
    mesh_triangles : TriangleArray
        array of triangles in mesh.
    mesh_triangles_colors : ShapeColorArray
        colors of triangles in mesh.
    vertices_index : IndexArray
        array with information which range of vertices belongs to which shape.
    mesh_triangles_index : IndexArray
        array with information which range of triangles belongs to which shape.
    mesh_vertices_index : IndexArray
        array with information which range of mesh vertices belongs to which shape.
    """

    z_index: ZOrderArray  # information about z-order of shapes
    vertices: CoordinateArray  # vertices of shapes

    mesh_vertices: CoordinateArray
    mesh_vertices_centers: CoordinateArray
    mesh_vertices_offsets: CoordinateArray

    mesh_triangles: TriangleArray
    mesh_triangles_colors: ShapeColorArray

    vertices_index: IndexArray  # offset of vertices for each shape
    mesh_triangles_index: (
        IndexArray  # offset of triangles and mesh vertices for each shape
    )
    mesh_vertices_index: IndexArray  # offset of mesh vertices for each shape


_SizeInformation = tuple[int, int, int, int, int]


def _ensure_color_arrays(shapes, face_colors=None, edge_colors=None):
    """Return as many face and edge colors as there are shapes in the input.

    Parameters
    ----------
    shapes : iterable of Shape
        Each Shape must be a subclass of Shape
    face_colors : iterable of face_color or None
        If None, default face colors will be used
    edge_colors : iterable of edge_color or None
        If None, default edge colors will be used

    Returns
    -------
    face_colors : np.ndarray
        Array of face colors
    edge_colors : np.ndarray
        Array of edge colors
    """
    if face_colors is None:
        face_colors = np.tile(
            np.array([1, 1, 1, 1], dtype=np.float32), (len(shapes), 1)
        )
    else:
        face_colors = np.asarray(face_colors, dtype=np.float32)

    if edge_colors is None:
        # default edge color is #777777, per napari.layers.Shapes constructor
        v = 7 / 15
        edge_colors = np.tile(
            np.array([v, v, v, 1], dtype=np.float32), (len(shapes), 1)
        )
    else:
        edge_colors = np.asarray(edge_colors, np.float32)

    if not len(face_colors) == len(edge_colors) == len(shapes):
        raise ValueError(
            trans._(
                'shapes, face_colors, and edge_colors must be the same length',
                deferred=True,
            )
        )

    return face_colors, edge_colors


def _calculate_array_sizes(shapes: Iterable[Shape]) -> _SizeInformation:
    """Calculate sizes needed for array preallocation.

    Parameters
    ----------
    shapes : iterable of Shape
        Each Shape must be a subclass of Shape

    Returns
    -------
    n_vertices : int
        Total number of vertices
    n_indices : int
        Total number of indices
    n_mesh_vertices : int
        Total number of mesh vertices
    n_face_tri : int
        Total number of face triangles
    n_edge_tri : int
        Total number of edge triangles
    """
    n_vertices = 0
    n_indices = 0
    n_mesh_vertices = 0
    n_face_tri = 0
    n_edge_tri = 0

    for shape in shapes:
        n_vertices += len(shape.data_displayed)
        n_indices += len(shape.data)
        n_mesh_vertices += len(shape._face_vertices) + len(
            shape._edge_vertices
        )
        n_face_tri += len(shape._face_triangles)
        n_edge_tri += len(shape._edge_triangles)

    return n_vertices, n_indices, n_mesh_vertices, n_face_tri, n_edge_tri


def _preallocate_arrays(
    shapes: Sequence[Shape], sizes: _SizeInformation
) -> MeshArrayDict:
    """Preallocate arrays for storing shape data.

    Parameters
    ----------
    shapes : Sequence of Shape
        Each Shape must be a subclass of Shape
    sizes : tuple
        Tuple containing sizes for preallocation:
        (n_vertices, n_indices, n_mesh_vertices,
        n_face_tri, n_edge_tri)

    Returns
    -------
    arrays : dict
        Dictionary containing preallocated arrays
    """
    n_shapes = len(shapes)
    n_vertices, _n_indices, n_mesh_vertices, n_face_tri, n_edge_tri = sizes

    # Determine the displayed dimensionality from the first shape
    # All shapes in a batch will have the same dimensionality
    dim: Literal[2, 3] = shapes[0]._face_vertices.shape[1]

    z_index = np.empty(n_shapes, dtype=np.int32)

    vertices_index = np.empty(n_shapes, dtype=IndexDtype)
    mesh_triangles_index = np.empty(n_shapes, dtype=IndexDtype)
    mesh_vertices_index = np.empty(n_shapes, dtype=IndexDtype)

    vertices = np.empty((n_vertices, dim), dtype=CoordinateDtype)

    mesh_vertices = np.empty((n_mesh_vertices, dim), dtype=CoordinateDtype)
    mesh_vertices_centers = np.empty(
        (n_mesh_vertices, dim), dtype=CoordinateDtype
    )
    mesh_vertices_offsets = np.empty(
        (n_mesh_vertices, dim), dtype=CoordinateDtype
    )

    total_triangles = n_face_tri + n_edge_tri
    mesh_triangles = np.empty((total_triangles, 3), dtype=TriangleDtype)
    mesh_triangles_colors = np.empty(
        (total_triangles, 4), dtype=ShapeColorDtype
    )

    return {
        'z_index': z_index,
        'vertices': vertices,  # type: ignore[typeddict-item]
        'mesh_vertices': mesh_vertices,  # type: ignore[typeddict-item]
        'mesh_vertices_centers': mesh_vertices_centers,  # type: ignore[typeddict-item]
        'mesh_vertices_offsets': mesh_vertices_offsets,  # type: ignore[typeddict-item]
        'mesh_triangles': mesh_triangles,  # type: ignore[typeddict-item]
        'mesh_triangles_colors': mesh_triangles_colors,  # type: ignore[typeddict-item]
        'vertices_index': vertices_index,
        'mesh_triangles_index': mesh_triangles_index,
        'mesh_vertices_index': mesh_vertices_index,
    }


def _fill_arrays(
    start_mesh_index: int,
    start_triangle_index: int,
    start_vertices_index: int,
    shapes: Iterable[Shape],
    face_colors: np.ndarray,
    edge_colors: np.ndarray,
    arrays: MeshArrayDict,
) -> None:
    """Fill pre-allocated arrays with shape data.

    Parameters
    ----------
    start_mesh_index : int
        The number of mesh vertices in the shape list before adding these
        shapes.
    start_triangle_index : int
        The number of mesh triangles in the shape list before adding these
    shapes : iterable of Shape
        Each Shape must be a subclass of Shape
    face_colors : np.ndarray
        Array of face colors
    edge_colors : np.ndarray
        Array of edge colors
    arrays : dict
        Dictionary containing preallocated arrays
    """
    z_index = arrays['z_index']
    vertices = arrays['vertices']
    # index = arrays['index']
    mesh_vertices = arrays['mesh_vertices']
    mesh_vertices_centers = arrays['mesh_vertices_centers']
    mesh_vertices_offsets = arrays['mesh_vertices_offsets']

    mesh_triangles = arrays['mesh_triangles']
    mesh_triangles_colors = arrays['mesh_triangles_colors']

    vertices_index = arrays['vertices_index']
    mesh_triangles_index = arrays['mesh_triangles_index']
    mesh_vertices_index = arrays['mesh_vertices_index']

    vertices_offset = 0
    mesh_vertices_offset = 0
    triangles_offset = 0

    for i, (shape, face_color, edge_color) in enumerate(
        zip(shapes, face_colors, edge_colors, strict=True)
    ):
        # Store z_index ("number of shapes" space)
        z_index[i] = shape.z_index
        vertices_ = shape.data_displayed
        n_vertices = len(vertices_)

        vertices_index[i] = start_vertices_index + vertices_offset + n_vertices
        mesh_triangles_index[i] = (
            start_triangle_index + triangles_offset + shape.triangles_count
        )
        mesh_vertices_index[i] = (
            start_mesh_index + mesh_vertices_offset + shape.vertices_count
        )

        # Store vertices data and update vertices offset
        vertices[vertices_offset : vertices_offset + n_vertices] = vertices_
        vertices_offset += n_vertices

        # Add faces to mesh
        face_vertices = shape._face_vertices
        n_face_vertices = len(face_vertices)

        face_vertices_slice = slice(
            mesh_vertices_offset, mesh_vertices_offset + n_face_vertices
        )

        mesh_vertices[face_vertices_slice] = face_vertices
        mesh_vertices_centers[face_vertices_slice] = face_vertices
        mesh_vertices_offsets[face_vertices_slice] = (
            0  # no shift for face vertices
        )

        # Store face triangles
        face_triangles = (
            shape._face_triangles + start_mesh_index + mesh_vertices_offset
        )
        n_face_triangles = len(face_triangles)
        face_triangles_slice = slice(
            triangles_offset, triangles_offset + n_face_triangles
        )
        mesh_triangles[face_triangles_slice] = face_triangles

        # Create and store face triangles colors
        mesh_triangles_colors[face_triangles_slice] = face_color

        # Update offsets
        mesh_vertices_offset += n_face_vertices
        triangles_offset += n_face_triangles

        # Add edges to mesh

        # Calculate edge vertices
        edge_vertices = shape._edge_vertices
        edge_offsets = shape._edge_offsets
        n_edge_vertices = len(edge_vertices)

        # Store edge vertices
        curr_vertices = edge_vertices + shape.edge_width * edge_offsets
        edge_vertices_slice = slice(
            mesh_vertices_offset, mesh_vertices_offset + n_edge_vertices
        )
        mesh_vertices[edge_vertices_slice] = curr_vertices
        mesh_vertices_centers[edge_vertices_slice] = edge_vertices
        mesh_vertices_offsets[edge_vertices_slice] = edge_offsets

        # Store edge triangles
        edge_triangles = (
            shape._edge_triangles + start_mesh_index + mesh_vertices_offset
        )
        n_edge_triangles = len(edge_triangles)
        edge_triangles_slice = slice(
            triangles_offset, triangles_offset + n_edge_triangles
        )
        mesh_triangles[edge_triangles_slice] = edge_triangles

        # Create and store edge triangles colors
        mesh_triangles_colors[edge_triangles_slice] = edge_color

        # Update offsets
        mesh_vertices_offset += n_edge_vertices
        triangles_offset += n_edge_triangles


def _batch_dec(meth):
    """
    Decorator to apply `self.batched_updates` to the current method.
    """

    @wraps(meth)
    def _wrapped(self, *args, **kwargs):
        with self.batched_updates():
            return meth(self, *args, **kwargs)

    return _wrapped


class ShapeList:
    """List of shapes class.

    Parameters
    ----------
    data : list
        List of Shape objects
    ndisplay : int
        Number of displayed dimensions.

    Attributes
    ----------
    shapes : (N, ) list
        Shape objects.
    data : (N, ) list of (M, D) array
        Data arrays for each shape.
    ndisplay : int
        Number of displayed dimensions.
    slice_keys : (N, 2, P) array
        Array of slice keys for each shape. Each slice key has the min and max
        values of the P non-displayed dimensions, useful for slicing
        multidimensional shapes. If the both min and max values of shape are
        equal then the shape is entirely contained within the slice specified
        by those values.
    shape_types : (N, ) list of str
        Name of shape type for each shape.
    edge_color : (N x 4) np.ndarray
        Array of RGBA edge colors for each shape.
    face_color : (N x 4) np.ndarray
        Array of RGBA face colors for each shape.
    edge_widths : (N, ) list of float
        Edge width for each shape.
    z_indices : (N, ) list of int
        z-index for each shape.

    Notes
    -----
    _vertices : np.ndarray
        MxN array of all displayed vertices from all shapes where N is equal to ndisplay
    _index : np.ndarray
        Length M array with the index (0, ..., N-1) of each shape that each
        vertex corresponds to
    _z_index : np.ndarray
        Length N array with z_index of each shape
    _z_order : np.ndarray
        Length N array with z_order of each shape. This must be a permutation
        of (0, ..., N-1).
    _mesh : Mesh
        Mesh object containing all the mesh information that will ultimately
        be rendered.
    """

    def __init__(
        self, data: typing.Iterable[Shape] = (), ndisplay: int = 2
    ) -> None:
        self._ndisplay = ndisplay
        self.shapes: list[Shape] = []
        self._displayed = np.array([])
        self._slice_key = np.array([])
        self.displayed_vertices = np.array([], dtype=CoordinateDtype)
        self.displayed_vertices_to_shape_num = np.array([], dtype=IndexDtype)
        self.displayed_indices = np.array([], dtype=IndexDtype)
        self._vertices = np.empty((0, self.ndisplay), dtype=CoordinateDtype)
        self._vertices_index: IndexArray = np.zeros(1, dtype=IndexDtype)
        self._z_index: IndexArray = np.empty(0, dtype=IndexDtype)
        self._z_order: IndexArray = np.empty(0, dtype=IndexDtype)

        self._mesh = Mesh(ndisplay=self.ndisplay)

        self._edge_color: ShapeColorArray = np.empty((0, 4))  # type: ignore[assignment]
        self._face_color: ShapeColorArray = np.empty((0, 4))  # type: ignore[assignment]

        # counter for the depth of re entrance of the context manager.
        self.__batched_level = 0
        self.__batch_force_call = False

        # Counter of number of time _update_displayed has been requested
        self.__update_displayed_called = 0
        if not isinstance(data, Sequence):
            data = list(data)
        self.add(data)

    def _vertices_slice(self, shape_index: int | np.integer) -> slice:
        """Return the slice of vertices for a given shape index."""
        start = self._vertices_index[shape_index]
        shape = self.shapes[shape_index]
        return slice(start, start + shape.data_displayed.shape[0])

    def _vertices_slice_available(self, shape_index: int) -> slice:
        """Return the available slice of vertices for a given shape index."""
        start = self._vertices_index[shape_index]
        end = self._vertices_index[shape_index + 1]
        return slice(start, end)

    def _mesh_vertices_slice(self, shape_index: int) -> slice:
        """Return the slice of mesh vertices for a given shape index."""
        start = self._mesh.vertices_index[shape_index]
        shape = self.shapes[shape_index]
        return slice(start, start + shape.vertices_count)

    def _mesh_vertices_slice_available(self, shape_index: int) -> slice:
        """Return the available slice of mesh vertices for a given shape index."""
        start = self._mesh.vertices_index[shape_index]
        end = self._mesh.vertices_index[shape_index + 1]
        return slice(start, end)

    def _mesh_triangles_slice(self, shape_index: int | np.integer) -> slice:
        """Return the slice of mesh triangles for a given shape index."""
        start = self._mesh.triangles_index[shape_index]
        shape = self.shapes[shape_index]
        return slice(start, start + shape.triangles_count)

    @staticmethod
    def _slicess_to_array(slices: Sequence[slice]) -> np.ndarray:
        count = sum(r.stop - r.start for r in slices)
        res = np.empty(count, dtype=np.int64)
        begin = 0
        for r in slices:
            end = begin + (r.stop - r.start)
            res[begin:end] = np.arange(r.start, r.stop)
            begin = end
        return res

    def _mesh_triangles_range_seq(
        self, shape_indexes: IndexArray
    ) -> np.ndarray | slice:
        """Return the range of mesh triangles for a sequence of shape indexes."""
        if (
            shape_indexes[-1] - shape_indexes[0] == len(shape_indexes) - 1
        ):  # If the sequence is continuous, return a range
            start = self._mesh.triangles_index[shape_indexes[0]]
            end = self._mesh.triangles_index[shape_indexes[-1]]
            end_shape = self.shapes[shape_indexes[-1]]
            return slice(start, end + end_shape.triangles_count)
        # If the sequence is not continuous, return a numpy array
        ranges = [self._mesh_triangles_slice(i) for i in shape_indexes]
        return self._slicess_to_array(ranges)

    def _vertices_range_seq(
        self, shape_indexes: IndexArray
    ) -> np.ndarray | slice:
        """Return the range of vertices for a sequence of shape indexes."""
        if (
            shape_indexes[-1] - shape_indexes[0] == len(shape_indexes) - 1
        ):  # If the sequence is continuous, return a range
            start = self._vertices_index[shape_indexes[0]]
            end = self._vertices_index[shape_indexes[-1]]
            end_shape = self.shapes[shape_indexes[-1]]
            return slice(start, end + end_shape.data_displayed.shape[0])
        # If the sequence is not continuous, return a numpy array
        slicess = [self._vertices_slice(i) for i in shape_indexes]
        return self._slicess_to_array(slicess)

    def _mesh_triangles_slice_available(self, shape_index: int) -> slice:
        """Return the available slice of mesh triangles for a given shape index."""
        start = self._mesh.triangles_index[shape_index]
        end = self._mesh.triangles_index[shape_index + 1]
        return slice(start, end)

    def _mesh_vertices_face_slice(self, shape_index: int) -> slice:
        """Return the slice of mesh vertices of face triangles for a given shape index."""
        shape = self.shapes[shape_index]
        start = self._mesh.vertices_index[shape_index]
        return slice(start, start + shape.face_vertices_count)

    def _mesh_triangles_face_slice(self, shape_index: int) -> slice:
        """Return the slice of mesh triangles of face triangles for a given shape index."""
        shape = self.shapes[shape_index]
        start = self._mesh.triangles_index[shape_index]
        return slice(start, start + shape.face_triangles_count)

    def _mesh_vertices_edge_slice(self, shape_index: int) -> slice:
        """Return the slice of mesh vertices of edge triangles for a given shape index."""
        shape = self.shapes[shape_index]
        start = (
            self._mesh.vertices_index[shape_index] + shape.face_vertices_count
        )
        return slice(start, start + shape.edge_vertices_count)

    def _mesh_triangles_edge_slice(self, shape_index: int) -> slice:
        """Return the slice of mesh triangles of edge triangles for a given shape index."""
        shape = self.shapes[shape_index]
        start = (
            self._mesh.triangles_index[shape_index]
            + shape.face_triangles_count
        )
        return slice(start, start + shape.edge_triangles_count)

    @contextmanager
    def batched_updates(self) -> Generator[None, None, None]:
        """
        Reentrant context manager to batch the display update

        `_update_displayed()` is called at _most_ once on exit of the context
        manager.

        There are two reason for this:

         1. Some updates are triggered by events, but sometimes multiple pieces
            of data that trigger events must be set before the data can be
            recomputed. For example changing number of dimension cause broacast
            error on partially update structures.
         2. Performance. Ideally we want to update the display only once.

        If no direct or indirect call to `_update_displayed()` are made inside
        the context manager, no the update logic is not called on exit.



        """
        assert self.__batched_level >= 0
        self.__batched_level += 1
        try:
            yield
        finally:
            if self.__batched_level == 1 and self.__update_displayed_called:
                self.__batch_force_call = True
                self._update_displayed()
                self.__batch_force_call = False
                self.__update_displayed_called = 0
            self.__batched_level -= 1

        assert self.__batched_level >= 0

    @property
    def data(self) -> list[npt.NDArray]:
        """list of (M, D) array: data arrays for each shape."""
        return [s.data for s in self.shapes]

    @property
    def ndisplay(self) -> int:
        """int: Number of displayed dimensions."""
        return self._ndisplay

    @ndisplay.setter
    def ndisplay(self, ndisplay: int) -> None:
        if self.ndisplay == ndisplay:
            return

        self._ndisplay = ndisplay
        self._mesh.ndisplay = self.ndisplay
        shapes = self.shapes
        face_color = self._face_color
        edge_color = self._edge_color

        with self.batched_updates():
            for shape in shapes:
                shape.ndisplay = self.ndisplay
            self.remove_all()
            self._add_multiple_shapes(
                shapes, face_colors=face_color, edge_colors=edge_color
            )

    @property
    def slice_keys(
        self,
    ) -> np.ndarray[tuple[int, Literal[2], int], np.dtype[np.int64]]:
        """(N, 2, P) array: slice key for each shape."""
        return np.array([s.slice_key for s in self.shapes])

    @property
    def shape_types(self) -> list[str]:
        """list of str: shape types for each shape."""
        return [s.name for s in self.shapes]

    @property
    def edge_color(self) -> ShapeColorArray:
        """(N x 4) np.ndarray: Array of RGBA edge colors for each shape"""
        return self._edge_color

    @edge_color.setter
    def edge_color(self, edge_color: npt.NDArray) -> None:
        self._set_color(edge_color, 'edge')

    @property
    def face_color(self) -> ShapeColorArray:
        """(N x 4) np.ndarray: Array of RGBA face colors for each shape"""
        return self._face_color

    @face_color.setter
    def face_color(self, face_color: npt.NDArray) -> None:
        self._set_color(face_color, 'face')

    @_batch_dec
    def _set_color(
        self, colors: npt.NDArray, attribute: Literal['edge', 'face']
    ) -> None:
        """Set the face_color or edge_color property

        Parameters
        ----------
        colors : (N, 4) np.ndarray
            The value for setting edge or face_color. There must
            be one color for each shape
        attribute : str in {'edge', 'face'}
            The name of the attribute to set the color of.
            Should be 'edge' for edge_color or 'face' for face_color.
        """
        n_shapes = len(self.data)
        if not np.array_equal(colors.shape, (n_shapes, 4)):
            raise ValueError(
                trans._(
                    '{attribute}_color must have shape ({n_shapes}, 4)',
                    deferred=True,
                    attribute=attribute,
                    n_shapes=n_shapes,
                )
            )

        update_method = getattr(self, f'update_{attribute}_colors')
        indices = np.arange(len(colors))
        update_method(indices, colors, update=False)
        self._update_displayed()

    @property
    def edge_widths(self) -> list[float]:
        """list of float: edge width for each shape."""
        return [s.edge_width for s in self.shapes]

    @property
    def z_indices(self) -> list[int]:
        """list of int: z-index for each shape."""
        return [s.z_index for s in self.shapes]

    @property
    def slice_key(self):
        """list: slice key for slicing n-dimensional shapes."""
        return self._slice_key

    @slice_key.setter
    @_batch_dec
    def slice_key(self, slice_key):
        slice_key = list(slice_key)
        if not np.array_equal(self._slice_key, slice_key):
            self._slice_key = slice_key
            self._clear_cache()
            self._update_displayed()

    def _update_displayed_triangles_to_shape_index(
        self, displayed_indices: IndexArray
    ) -> None:
        """Update the displayed triangles to shape index mapping."""
        if (
            self._mesh.displayed_triangles_to_shape_index.shape[0]
            != self._mesh.displayed_triangles.shape[0]
        ):
            self._mesh.displayed_triangles_to_shape_index = np.full(
                self._mesh.displayed_triangles.shape[0], -1, dtype=IndexDtype
            )
        shift_idx = 0
        for i in displayed_indices:
            begin = self._mesh.triangles_index[i]
            end = self._mesh.triangles_index[i + 1]
            elem_num = end - begin
            self._mesh.displayed_triangles_to_shape_index[
                shift_idx : shift_idx + elem_num
            ] = i
            shift_idx += elem_num

    def _update_displayed_vertices_to_shape_num(
        self, displayed_indices: IndexArray
    ) -> None:
        """Update the displayed vertices to shape index mapping."""
        if (
            self.displayed_vertices_to_shape_num.shape[0]
            != self.displayed_vertices.shape[0]
        ):
            self.displayed_vertices_to_shape_num = np.full(
                self.displayed_vertices.shape[0], -1, dtype=IndexDtype
            )
        shift_idx = 0
        for i in displayed_indices:
            begin = self._vertices_index[i]
            end = self._vertices_index[i + 1]
            elem_num = end - begin
            self.displayed_vertices_to_shape_num[
                shift_idx : shift_idx + elem_num
            ] = i
            shift_idx += elem_num

    def _update_displayed(self) -> None:
        """Update the displayed data based on the slice key.

        This method must be called from within the `batched_updates` context
        manager:
        """
        assert self.__batched_level >= 1, (
            'call _update_displayed from within self.batched_updates context manager'
        )
        if not self.__batch_force_call:
            self.__update_displayed_called += 1
            return

        # The list slice key is repeated to check against both the min and
        # max values stored in the shapes slice key.
        slice_key = np.array([self.slice_key, self.slice_key])

        # Slice key must exactly match mins and maxs of shape as then the
        # shape is entirely contained within the current slice.
        if len(self.shapes) > 0:
            self._displayed = np.all(
                np.abs(self.slice_keys - slice_key) < 0.5, axis=(1, 2)
            )
        else:
            self._displayed = np.array([])
        disp_indices: IndexArray = np.nonzero(self._displayed)[0]  # type: ignore[assignment]

        z_order = self._mesh.triangles_z_order

        triangle_ranges: IndexArray | slice
        vertices_range: IndexArray | slice

        if disp_indices.size == 0:
            triangle_ranges = np.array([], dtype=np.int64)
            vertices_range = np.array([], dtype=np.int64)
        else:
            triangle_ranges = self._mesh_triangles_range_seq(disp_indices)
            vertices_range = self._vertices_range_seq(disp_indices)

        self._mesh.displayed_triangles = self._mesh.triangles[
            z_order[triangle_ranges]
        ]

        self._update_displayed_triangles_to_shape_index(disp_indices)

        self._mesh.displayed_triangles_colors = self._mesh.triangles_colors[
            z_order[triangle_ranges]
        ]

        self.displayed_vertices = self._vertices[vertices_range]
        self._update_displayed_vertices_to_shape_num(disp_indices)
        self.displayed_index = self._vertices_index[disp_indices]

    def add(
        self,
        shape: Shape | Sequence[Shape],
        face_color=None,
        edge_color=None,
        shape_index=None,
        z_refresh=True,
    ):
        """Adds a single Shape object (single add mode) or multiple Shapes (multiple shape mode, which is much faster)

        If shape is a single instance of subclass Shape then single add mode will be used, otherwise multiple add mode

        Parameters
        ----------
        shape : single Shape or iterable of Shape
            Each shape must be a subclass of Shape, one of "{'Line', 'Rectangle',
            'Ellipse', 'Path', 'Polygon'}"
        face_color : color (or iterable of colors of same length as shape)
        edge_color : color (or iterable of colors of same length as shape)
        shape_index : None | int
            If int then edits the shape date at current index. To be used in
            conjunction with `remove` when renumber is `False`. If None, then
            appends a new shape to end of shapes list
            Must be None in multiple shape mode.
        z_refresh : bool
            If set to true, the mesh elements are reindexed with the new z order.
            When shape_index is provided, z_refresh will be overwritten to false,
            as the z indices will not change.
            When adding a batch of shapes, set to false  and then call
            ShapesList._update_z_order() once at the end.
        """
        # single shape mode
        if issubclass(type(shape), Shape):
            self._add_single_shape(
                shape=shape,
                face_color=face_color,
                edge_color=edge_color,
                shape_index=shape_index,
                z_refresh=z_refresh,
            )
        # multiple shape mode
        elif isinstance(shape, Iterable):
            if shape_index is not None:
                raise ValueError(
                    trans._(
                        'shape_index must be None when adding multiple shapes',
                        deferred=True,
                    )
                )
            self._add_multiple_shapes(
                shapes=shape,
                face_colors=face_color,
                edge_colors=edge_color,
                z_refresh=z_refresh,
            )
        else:
            raise TypeError(
                trans._(
                    'Cannot add single nor multiple shape',
                    deferred=True,
                )
            )

    def _add_single_shape(
        self,
        shape,
        face_color=None,
        edge_color=None,
        shape_index=None,
        z_refresh=True,
    ):
        """Adds a single Shape object

        Parameters
        ----------
        shape : subclass Shape
            Must be a subclass of Shape, one of "{'Line', 'Rectangle',
            'Ellipse', 'Path', 'Polygon'}"
        shape_index : None | int
            If int then edits the shape date at current index. To be used in
            conjunction with `remove` when renumber is `False`. If None, then
            appends a new shape to end of shapes list
        z_refresh : bool
            If set to true, the mesh elements are reindexed with the new z order.
            When shape_index is provided, z_refresh will be overwritten to false,
            as the z indices will not change.
            When adding a batch of shapes, set to false  and then call
            ShapesList._update_z_order() once at the end.
        """
        if not issubclass(type(shape), Shape):
            raise TypeError(
                trans._(
                    'shape must be subclass of Shape',
                    deferred=True,
                )
            )

        if shape_index is None:
            self.shapes.append(shape)
            self._z_index = np.append(self._z_index, shape.z_index)

            if face_color is None:
                face_color = np.array([1, 1, 1, 1])
            self._face_color = np.vstack([self._face_color, face_color])
            if edge_color is None:
                edge_color = np.array([0, 0, 0, 1])
            self._edge_color = np.vstack([self._edge_color, edge_color])
        else:
            z_refresh = False
            self.shapes[shape_index] = shape
            self._z_index[shape_index] = shape.z_index

            if face_color is None:
                face_color = self._face_color[shape_index]
            else:
                self._face_color[shape_index, :] = face_color
            if edge_color is None:
                edge_color = self._edge_color[shape_index]
            else:
                self._edge_color[shape_index, :] = edge_color

        vertices_ = shape.data_displayed
        self._vertices_index = np.append(
            self._vertices_index,
            [len(self._vertices) + len(vertices_)],
            axis=0,
        )

        self._vertices = np.append(self._vertices, vertices_, axis=0)

        # Add faces to mesh
        m = len(self._mesh.vertices)
        vertices = shape._face_vertices
        self._mesh.vertices_index = np.append(
            self._mesh.vertices_index,
            [len(self._mesh.vertices) + shape.vertices_count],
            axis=0,
        )
        self._mesh.vertices = np.append(self._mesh.vertices, vertices, axis=0)
        vertices = shape._face_vertices
        self._mesh.vertices_centers = np.append(
            self._mesh.vertices_centers, vertices, axis=0
        )
        vertices = np.zeros(shape._face_vertices.shape)
        self._mesh.vertices_offsets = np.append(
            self._mesh.vertices_offsets, vertices, axis=0
        )

        triangles = shape._face_triangles + m
        self._mesh.triangles_index = np.append(
            self._mesh.triangles_index,
            [len(self._mesh.triangles) + shape.triangles_count],
            axis=0,
        )
        self._mesh.triangles = np.append(
            self._mesh.triangles, triangles, axis=0
        )
        color_array = np.repeat([face_color], len(triangles), axis=0)
        self._mesh.triangles_colors = np.append(
            self._mesh.triangles_colors, color_array, axis=0
        )

        # Add edges to mesh
        m = len(self._mesh.vertices)
        vertices = (
            shape._edge_vertices + shape.edge_width * shape._edge_offsets
        )
        self._mesh.vertices = np.append(self._mesh.vertices, vertices, axis=0)
        vertices = shape._edge_vertices
        self._mesh.vertices_centers = np.append(
            self._mesh.vertices_centers, vertices, axis=0
        )
        vertices = shape._edge_offsets
        self._mesh.vertices_offsets = np.append(
            self._mesh.vertices_offsets, vertices, axis=0
        )

        triangles = shape._edge_triangles + m
        self._mesh.triangles = np.append(
            self._mesh.triangles, triangles, axis=0
        )
        color_array = np.repeat([edge_color], len(triangles), axis=0)
        self._mesh.triangles_colors = np.append(
            self._mesh.triangles_colors, color_array, axis=0
        )

        if z_refresh:
            # Set z_order
            self._update_z_order()
        self._clear_cache()

    def _extend_meshes(self, face_colors, edge_colors, arrays: MeshArrayDict):
        """Assemble mesh properties from filled arrays.

        Parameters
        ----------
        face_colors : np.ndarray
            Array of face colors
        edge_colors : np.ndarray
            Array of edge colors
        arrays : dict
            Dictionary containing filled arrays
        """
        z_index = arrays['z_index']
        vertices = arrays['vertices']

        mesh_vertices = arrays['mesh_vertices']
        mesh_vertices_centers = arrays['mesh_vertices_centers']
        mesh_vertices_offsets = arrays['mesh_vertices_offsets']

        mesh_triangles = arrays['mesh_triangles']
        mesh_triangles_colors = arrays['mesh_triangles_colors']

        vertices_index = arrays['vertices_index']
        mesh_triangles_index = arrays['mesh_triangles_index']
        mesh_vertices_index = arrays['mesh_vertices_index']

        # assemble properties
        self._z_index = np.append(self._z_index, z_index, axis=0)
        self._face_color = np.vstack((self._face_color, face_colors))
        self._edge_color = np.vstack((self._edge_color, edge_colors))
        self._vertices = np.vstack((self._vertices, vertices))
        self._vertices_index = np.append(
            self._vertices_index, vertices_index, axis=0
        )

        self._mesh.vertices = np.vstack((self._mesh.vertices, mesh_vertices))
        self._mesh.vertices_centers = np.vstack(
            (self._mesh.vertices_centers, mesh_vertices_centers)
        )
        self._mesh.vertices_offsets = np.vstack(
            (self._mesh.vertices_offsets, mesh_vertices_offsets)
        )
        self._mesh.vertices_index = np.append(
            self._mesh.vertices_index, mesh_vertices_index, axis=0
        )

        self._mesh.triangles = np.vstack(
            (self._mesh.triangles, mesh_triangles)
        )
        self._mesh.triangles_index = np.append(
            self._mesh.triangles_index, mesh_triangles_index, axis=0
        )
        self._mesh.triangles_colors = np.vstack(
            (self._mesh.triangles_colors, mesh_triangles_colors)
        )

    def _add_multiple_shapes(
        self,
        shapes: Sequence[Shape],
        face_colors=None,
        edge_colors=None,
        z_refresh=True,
    ) -> None:
        """Add multiple shapes at once (faster than adding them one by one)

        Parameters
        ----------
        shapes : iterable of Shape
            Each Shape must be a subclass of Shape, one of "{'Line', 'Rectangle',
            'Ellipse', 'Path', 'Polygon'}"
        face_colors : iterable of face_color
        edge_colors : iterable of edge_color
        z_refresh : bool
            If set to true, the mesh elements are reindexed with the new z order.
            When shape_index is provided, z_refresh will be overwritten to false,
            as the z indices will not change.
            When adding a batch of shapes, set to false  and then call
            ShapesList._update_z_order() once at the end.

        TODO: Currently shares a lot of code with `add()`, with the
        difference being that `add()` supports inserting shapes at a specific
        `shape_index`, whereas `add_multiple` will append them as a full batch
        """
        if len(shapes) == 0:
            return

        # Validate inputs and prepare colors
        if not all(issubclass(type(shape), Shape) for shape in shapes):
            raise ValueError(
                trans._(
                    'all shapes must be subclass of Shape',
                    deferred=True,
                )
            )
        face_colors, edge_colors = _ensure_color_arrays(
            shapes, face_colors, edge_colors
        )

        # Calculate sizes for preallocation
        sizes = _calculate_array_sizes(shapes)

        # Preallocate arrays
        arrays = _preallocate_arrays(shapes, sizes)

        # Fill pre-allocated arrays with mesh and index data
        _fill_arrays(
            len(self._mesh.vertices),
            len(self._mesh.triangles),
            len(self._vertices),
            shapes,
            face_colors,
            edge_colors,
            arrays,
        )

        # Update local arrays appending mesh properties
        self._extend_meshes(face_colors, edge_colors, arrays)

        # Update list of shapes
        self.shapes.extend(shapes)

        if z_refresh:
            # Set z_order
            self._update_z_order()
        self._clear_cache()

    @_batch_dec
    def remove_all(self):
        """Removes all shapes"""
        self.shapes = []
        self._vertices = np.empty((0, self.ndisplay))  # type: ignore[assignment]
        self._vertices_index = np.zeros(1, dtype=IndexDtype)
        self._z_index = np.empty(0, dtype=IndexDtype)
        self._z_order = np.empty(0, dtype=ZOrderDtype)
        self._edge_color = np.empty((0, 4), dtype=ShapeColorDtype)  # type: ignore[assignment]
        self._face_color = np.empty((0, 4), dtype=ShapeColorDtype)  # type: ignore[assignment]
        self._mesh.clear()
        self._update_displayed()

    @_batch_dec
    def update(self, index: int) -> None:
        """update shape at index `index`"""
        self._update_vertices(index)
        self._update_mesh_triangles(index)
        self._update_mesh_vertices(index, edge=True, face=True)
        self._update_displayed()

    def _update_vertices(self, index: int) -> None:
        shape = self.shapes[index]
        vertices_slice = self._vertices_slice_available(index)
        curr_vert_count = vertices_slice.stop - vertices_slice.start
        if shape.data_displayed.shape[0] == curr_vert_count:
            # If the number of vertices is the same, just update the data
            self._vertices[vertices_slice] = shape.data_displayed
        elif shape.data_displayed.shape[0] < curr_vert_count:
            # To avoid relocation, we add first point few times for padding.
            new_slice = slice(
                vertices_slice.start,
                vertices_slice.start + shape.data_displayed.shape[0],
            )
            padding_slice = slice(new_slice.stop, vertices_slice.stop)
            self._vertices[new_slice] = shape.data_displayed
            self._vertices[padding_slice] = shape.data_displayed[0]
        else:
            # there are more vertices in the shape than in the mesh
            before_array = self._vertices[: vertices_slice.start]
            after_array = self._vertices[vertices_slice.stop :]
            self._vertices = np.concatenate(
                [before_array, shape.data_displayed, after_array]
            )
            self._vertices_index[index + 1 :] += (
                shape.data_displayed.shape[0] - curr_vert_count
            )

    def _update_mesh_triangles(self, index: int) -> None:
        """Updates the mesh triangle data for a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        """
        shape = self.shapes[index]
        if index == 0:
            triangle_shift = 0
        else:
            triangle_shift = self._mesh_vertices_slice_available(
                index - 1
            ).stop
        triangles_slice = self._mesh_triangles_slice_available(index)
        current_triangles_count = triangles_slice.stop - triangles_slice.start
        new_triangle_count = shape.triangles_count
        prev_vertices_slice = self._mesh_vertices_slice_available(index)
        prev_vertices_count = (
            prev_vertices_slice.stop - prev_vertices_slice.start
        )
        new_vertices_count = shape.vertices_count
        if new_triangle_count <= current_triangles_count:
            face_slice = slice(
                triangles_slice.start,
                triangles_slice.start + shape.face_triangles_count,
            )
            edge_slice = slice(
                triangles_slice.start + shape.face_triangles_count,
                triangles_slice.start + shape.triangles_count,
            )
            self._mesh.triangles[face_slice] = (
                shape._face_triangles + triangle_shift
            )
            self._mesh.triangles_colors[face_slice] = self._face_color[index]
            self._mesh.triangles[edge_slice] = shape._edge_triangles + (
                triangle_shift + shape.face_vertices_count
            )
            self._mesh.triangles_colors[edge_slice] = self._edge_color[index]
            if new_triangle_count < current_triangles_count:
                padding_slice = slice(
                    triangles_slice.start + shape.triangles_count,
                    triangles_slice.stop,
                )
                self._mesh.triangles[padding_slice] = triangle_shift
            if new_vertices_count > prev_vertices_count:
                # Shift triangles indices if more vertices were added
                self._mesh.triangles[triangles_slice.stop :] += (
                    new_vertices_count - prev_vertices_count
                )
        else:
            # there are more triangles in the shape than in the mesh
            before_array = self._mesh.triangles[: triangles_slice.start]
            after_array = self._mesh.triangles[triangles_slice.stop :]
            after_array += new_vertices_count - prev_vertices_count
            self._mesh.triangles = np.concatenate(
                [
                    before_array,
                    shape._face_triangles + triangle_shift,
                    shape._edge_triangles
                    + (triangle_shift + shape.face_vertices_count),
                    after_array,
                ]
            )
            self._mesh.triangles_index[index + 1 :] += (
                new_triangle_count - current_triangles_count
            )
            # Update triangles colors
            self._mesh.triangles_colors = np.concatenate(
                [
                    self._mesh.triangles_colors[: triangles_slice.start],
                    np.repeat(
                        [self._face_color[index]],
                        shape.face_triangles_count,
                        axis=0,
                    ),
                    np.repeat(
                        [self._edge_color[index]],
                        shape.edge_triangles_count,
                        axis=0,
                    ),
                    self._mesh.triangles_colors[triangles_slice.stop :],
                ]
            )
            self._update_z_order()

    def remove(self, index: int, renumber: bool = True) -> None:
        """Removes a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be removed.
        renumber : bool
            Bool to indicate whether to renumber all shapes or not. If not the
            expectation is that this shape is being immediately added back to the
            list using `add_shape`.
        """
        self.remove_multiple([index], renumber=renumber)

    @_batch_dec
    def remove_multiple(self, indices: list[int], renumber: bool = True):
        """Removes multiple shapes located at indices.

        Parameters
        ----------
        indices : list of int
            Locations in list of the shapes to be removed. Assumed to be sorted descending.
        renumber : bool
            Bool to indicate whether to renumber all shapes or not.
        """
        if not indices:
            return

        # Remove indices
        vert_slices = [self._vertices_slice_available(i) for i in indices]
        vert_indices_to_del = np.concatenate(
            [np.arange(s.start, s.stop) for s in vert_slices]
        )
        self._vertices = np.delete(self._vertices, vert_indices_to_del, axis=0)

        vert_counts = np.diff(self._vertices_index)
        new_vert_counts = np.delete(vert_counts, indices)
        self._vertices_index = np.concatenate(
            ([0], np.cumsum(new_vert_counts))
        ).astype(IndexDtype)

        # Remove vertices
        mesh_vert_slices = [
            self._mesh_vertices_slice_available(i) for i in indices
        ]
        mesh_vert_indices_to_del = np.concatenate(
            [np.arange(s.start, s.stop) for s in mesh_vert_slices]
        )
        # Get the shift for triangles caused by the removal of mesh vertices
        deleted_vertex_shift = np.zeros(len(self._mesh.vertices), dtype=int)
        deleted_vertex_shift[mesh_vert_indices_to_del] = 1
        deleted_vertex_shift = np.cumsum(deleted_vertex_shift)
        self._mesh.vertices = np.delete(
            self._mesh.vertices, mesh_vert_indices_to_del, axis=0
        )
        self._mesh.vertices_centers = np.delete(
            self._mesh.vertices_centers, mesh_vert_indices_to_del, axis=0
        )
        self._mesh.vertices_offsets = np.delete(
            self._mesh.vertices_offsets, mesh_vert_indices_to_del, axis=0
        )

        mesh_vert_counts = np.diff(self._mesh.vertices_index)
        new_mesh_vert_counts = np.delete(mesh_vert_counts, indices)
        self._mesh.vertices_index = np.concatenate(
            ([0], np.cumsum(new_mesh_vert_counts))
        ).astype(IndexDtype)

        # Remove triangles
        mesh_tri_slices = [
            self._mesh_triangles_slice_available(i) for i in indices
        ]
        mesh_tri_indices_to_del = np.concatenate(
            [np.arange(s.start, s.stop) for s in mesh_tri_slices]
        )
        self._mesh.triangles -= deleted_vertex_shift[self._mesh.triangles]
        self._mesh.triangles = np.delete(
            self._mesh.triangles, mesh_tri_indices_to_del, axis=0
        )
        self._mesh.triangles_colors = np.delete(
            self._mesh.triangles_colors, mesh_tri_indices_to_del, axis=0
        )

        mesh_tri_counts = np.diff(self._mesh.triangles_index)
        new_mesh_tri_counts = np.delete(mesh_tri_counts, indices)
        self._mesh.triangles_index = np.concatenate(
            ([0], np.cumsum(new_mesh_tri_counts))
        ).astype(IndexDtype)

        if renumber:
            for i in indices:
                del self.shapes[i]
            self._z_index = np.delete(self._z_index, indices)
            self._update_z_order()

        self._clear_cache()

    @_batch_dec
    def _update_mesh_vertices(self, index, edge=False, face=False):
        """Updates the mesh vertex data and vertex data for a single shape
        located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        edge : bool
            Bool to indicate whether to update mesh vertices corresponding to
            edges
        face : bool
            Bool to indicate whether to update mesh vertices corresponding to
            faces and to update the underlying shape vertices
        """
        shape = self.shapes[index]
        if edge and face:
            shape_slice = self._mesh_vertices_slice_available(index)
            current_range = shape_slice.stop - shape_slice.start
            if current_range < shape.vertices_count:
                # account for edge width
                edge_vertices_with_width = (
                    shape._edge_vertices
                    + shape.edge_width * shape._edge_offsets
                )
                # need to allocate_more space
                self._mesh.vertices = np.concatenate(
                    [
                        self._mesh.vertices[: shape_slice.start],
                        shape._face_vertices,
                        edge_vertices_with_width,
                        self._mesh.vertices[shape_slice.stop :],
                    ]
                )
                self._mesh.vertices_centers = np.concatenate(
                    [
                        self._mesh.vertices_centers[: shape_slice.start],
                        shape._face_vertices,
                        shape._edge_vertices,
                        self._mesh.vertices_centers[shape_slice.stop :],
                    ]
                )
                self._mesh.vertices_offsets = np.concatenate(
                    [
                        self._mesh.vertices_offsets[: shape_slice.start],
                        np.zeros(
                            shape._face_vertices.shape,
                            dtype=shape._face_vertices.dtype,
                        ),
                        shape._edge_offsets,
                        self._mesh.vertices_offsets[shape_slice.stop :],
                    ]
                )
                diff = shape.vertices_count - current_range
                self._mesh.vertices_index[index + 1 :] += diff
                return
            if current_range > shape.vertices_count:
                zeros_slice = slice(
                    shape_slice.start + shape.vertices_count, shape_slice.stop
                )
                self._mesh.vertices[zeros_slice] = 0
                self._mesh.vertices_centers[zeros_slice] = 0
                self._mesh.vertices_offsets[zeros_slice] = 0

        if edge:
            indices = self._mesh_vertices_edge_slice(index)
            self._mesh.vertices[indices] = (
                shape._edge_vertices + shape.edge_width * shape._edge_offsets
            )
            self._mesh.vertices_centers[indices] = shape._edge_vertices
            self._mesh.vertices_offsets[indices] = shape._edge_offsets
            self._update_displayed()

        if face:
            indices = self._mesh_vertices_face_slice(index)
            self._mesh.vertices[indices] = shape._face_vertices
            self._mesh.vertices_centers[indices] = shape._face_vertices
            indices = self._vertices_slice(index)
            self._vertices[indices] = shape.data_displayed
            self._update_displayed()
        self._clear_cache()

    @_batch_dec
    def _update_z_order(self):
        """Updates the z order of the triangles given the z_index list"""
        self._z_order = np.argsort(self._z_index, kind='stable')  # type: ignore[assignment]
        if len(self._z_order) == 0:
            self._mesh.triangles_z_order = np.empty(0, dtype=ZOrderDtype)
        else:
            idx = self._mesh.triangles_index
            counts = np.empty(idx.shape, dtype=idx.dtype)
            counts[:-1] = idx[1:] - idx[:-1]
            counts[-1] = len(self._mesh.triangles) - idx[-1]
            triangles_z_order = [
                np.arange(idx[z], idx[z] + counts[z]) for z in self._z_order
            ]
            self._mesh.triangles_z_order = np.concatenate(triangles_z_order)
        self._update_displayed()

    def edit(
        self, index, data, face_color=None, edge_color=None, new_type=None
    ):
        """Updates the data of a single shape located at index. If
        `new_type` is not None then converts the shape type to the new type

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        data : np.ndarray
            NxD array of vertices.
        new_type : None | str | Shape
            If string , must be one of "{'line', 'rectangle', 'ellipse',
            'path', 'polygon'}".
        """
        if new_type is not None:
            cur_shape = self.shapes[index]
            if isinstance(new_type, str):
                shape_type = ShapeType(new_type)
                if shape_type in shape_classes:
                    shape_cls = shape_classes[shape_type]
                else:
                    raise ValueError(
                        trans._(
                            '{shape_type} must be one of {shape_classes}',
                            deferred=True,
                            shape_type=shape_type,
                            shape_classes=set(shape_classes),
                        )
                    )
            else:
                shape_cls = new_type
            shape = shape_cls(
                data,
                edge_width=cur_shape.edge_width,
                z_index=cur_shape.z_index,
                dims_order=cur_shape.dims_order,
            )
            self.shapes[index] = shape
        else:
            shape = self.shapes[index]
            shape.data = data

        if face_color is not None:
            self._face_color[index] = face_color
        if edge_color is not None:
            self._edge_color[index] = edge_color

        self.update(index)
        self._update_z_order()

    def update_edge_width(self, index, edge_width):
        """Updates the edge width of a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        edge_width : float
            thickness of lines and edges.
        """
        self.shapes[index].edge_width = edge_width
        self._update_mesh_vertices(index, edge=True)

    @_batch_dec
    def update_edge_color(
        self, index: int, edge_color: ShapeColor, update: bool = True
    ) -> None:
        """Updates the edge color of a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        edge_color : np.ndarray of shape (4,)
            Array of RGBA values for the edge color.
        update : bool
            If True, update the mesh with the new color property. Set to False to avoid
            repeated updates when modifying multiple shapes. Default is True.
        """
        self._edge_color[index] = edge_color
        indices = self._mesh_triangles_edge_slice(index)
        self._mesh.triangles_colors[indices] = self._edge_color[index]
        if update:
            self._update_displayed()

    @_batch_dec
    def update_edge_colors(
        self,
        indices: Iterable[int],
        edge_colors: ShapeColor | ShapeColorArray,
        update: bool = True,
    ) -> None:
        """same as update_edge_color() but for multiple indices/edgecolors at once"""
        edge_colors_: Iterable[ShapeColor]
        if edge_colors.ndim == 1:
            edge_colors_ = repeat(typing.cast(ShapeColor, edge_colors))
        elif edge_colors.ndim == 2 and edge_colors.shape[0] == 1:
            edge_colors_ = repeat(typing.cast(ShapeColor, edge_colors[0]))
        else:
            edge_colors_ = edge_colors
        for i, color in zip(indices, edge_colors_, strict=False):
            self.update_edge_color(i, color, update=False)
        if update:
            self._update_displayed()

    @_batch_dec
    def update_face_color(
        self, index: int, face_color: ShapeColor, update: bool = True
    ) -> None:
        """Updates the face color of a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        face_color : np.ndarray of shape (4,)
            Array of RGBA values for the face color.
        update : bool
            If True, update the mesh with the new color property. Set to False to avoid
            repeated updates when modifying multiple shapes. Default is True.
        """
        self._face_color[index] = face_color
        indices = self._mesh_triangles_face_slice(index)
        self._mesh.triangles_colors[indices] = face_color
        if update:
            self._update_displayed()

    @_batch_dec
    def update_face_colors(
        self,
        indices: Iterable[int],
        face_colors: ShapeColor | ShapeColorArray,
        update: bool = True,
    ) -> None:
        """same as update_face_color() but for multiple indices/facecolors at once"""
        face_colors_: Iterable[ShapeColor]
        if face_colors.ndim == 1:
            face_colors_ = repeat(typing.cast(ShapeColor, face_colors))
        elif face_colors.ndim == 2 and face_colors.shape[0] == 1:
            face_colors_ = repeat(typing.cast(ShapeColor, face_colors[0]))
        else:
            face_colors_ = face_colors

        for i, color in zip(indices, face_colors_, strict=False):
            self.update_face_color(i, color, update=False)
        if update:
            self._update_displayed()

    def update_dims_order(self, dims_order):
        """Updates dimensions order for all shapes.

        Parameters
        ----------
        dims_order : (D,) list
            Order that the dimensions are rendered in.
        """
        for index in range(len(self.shapes)):
            if self.shapes[index].dims_order != dims_order:
                shape = self.shapes[index]
                shape.dims_order = dims_order
                self.update(index)
        self._update_z_order()

    def update_z_index(self, index, z_index):
        """Updates the z order of a single shape located at index.

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        z_index : int
            Specifier of z order priority. Shapes with higher z order are
            displayed ontop of others.
        """
        self.shapes[index].z_index = z_index
        self._z_index[index] = z_index
        self._update_z_order()

    def shift(self, index, shift):
        """Performs a 2D shift on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        shift : np.ndarray
            length 2 array specifying shift of shapes.
        """
        self.shapes[index].shift(shift)
        self._update_mesh_vertices(index, edge=True, face=True)

    def scale(self, index, scale, center=None):
        """Performs a scaling on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        scale : float, list
            scalar or list specifying rescaling of shape.
        center : list
            length 2 list specifying coordinate of center of scaling.
        """
        self.shapes[index].scale(scale, center=center)
        self.update(index)
        self._update_z_order()

    def rotate(self, index, angle, center=None):
        """Performs a rotation on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        angle : float
            angle specifying rotation of shape in degrees.
        center : list
            length 2 list specifying coordinate of center of rotation.
        """
        self.shapes[index].rotate(angle, center=center)
        self._update_mesh_vertices(index, edge=True, face=True)

    def flip(self, index, axis, center=None):
        """Performs an vertical flip on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        axis : int
            integer specifying axis of flip. `0` flips horizontal, `1` flips
            vertical.
        center : list
            length 2 list specifying coordinate of center of flip axes.
        """
        self.shapes[index].flip(axis, center=center)
        self._update_mesh_vertices(index, edge=True, face=True)

    def transform(self, index, transform):
        """Performs a linear transform on a single shape located at index

        Parameters
        ----------
        index : int
            Location in list of the shape to be changed.
        transform : np.ndarray
            2x2 array specifying linear transform.
        """
        self.shapes[index].transform(transform)
        self.update(index)
        self._update_z_order()
        self._clear_cache()

    def outline(
        self, indices: int | Sequence[int]
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Finds outlines of shapes listed in indices

        Parameters
        ----------
        indices : int | Sequence[int]
            Location in list of the shapes to be outline.
            If sequence, all elements should be ints

        Returns
        -------
        centers : np.ndarray
            Nx2 array of centers of outline
        offsets : np.ndarray
            Nx2 array of offsets of outline
        triangles : np.ndarray
            Mx3 array of any indices of vertices for triangles of outline
        """
        if isinstance(indices, Sequence) and len(indices) == 1:
            indices = indices[0]
        if not isinstance(indices, Sequence):
            shape = self.shapes[indices]
            return (
                shape._edge_vertices,
                shape._edge_offsets,
                shape._edge_triangles,
            )
        return self.outlines(indices)

    def outlines(
        self, indices: Sequence[int]
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Finds outlines of shapes listed in indices, using chunked processing.

        Parameters
        ----------
        indices : Sequence[int]
            Location in list of the shapes to be outline.

        Returns
        -------
        centers : np.ndarray
            Nx2 array of centers of outline
        offsets : np.ndarray
            Nx2 array of offsets of outline
        triangles : np.ndarray
            Mx3 array of any indices of vertices for triangles of outline
        """
        # Based on benchmarking, a chunk_size of 500 provides a good balance
        # of performance for a wide range of shape counts.
        chunk_size = 500

        if not indices:
            return (
                np.empty((0, self.ndisplay), dtype=CoordinateDtype),
                np.empty((0, self.ndisplay), dtype=CoordinateDtype),
                np.empty((0, 3), dtype=TriangleDtype),
            )

        centers_blocks = []
        offsets_blocks = []
        triangles_blocks = []
        n_verts_cumsum = 0

        for start in range(0, len(indices), chunk_size):
            chunk_indices = indices[start : start + chunk_size]
            chunk_shapes = [self.shapes[i] for i in chunk_indices]

            chunk_centers = []
            chunk_offsets = []
            chunk_tris = []
            n_verts_per_shape = []
            n_tris_per_shape = []

            for s in chunk_shapes:
                verts = s._edge_vertices
                chunk_centers.append(verts)
                chunk_offsets.append(s._edge_offsets)
                tris = s._edge_triangles
                chunk_tris.append(tris)
                n_verts_per_shape.append(verts.shape[0])
                n_tris_per_shape.append(tris.shape[0])

            centers = np.concatenate(chunk_centers)
            offsets = np.concatenate(chunk_offsets)
            triangles = np.concatenate(chunk_tris)

            # Offset triangle indices within chunk and across blocks
            if triangles.size > 0:
                vert_offsets = np.zeros(
                    len(n_verts_per_shape), dtype=triangles.dtype
                )
                if len(vert_offsets) > 1:
                    np.cumsum(n_verts_per_shape[:-1], out=vert_offsets[1:])
                tri_offsets = np.repeat(vert_offsets, n_tris_per_shape)
                triangles = (
                    triangles + tri_offsets[:, np.newaxis] + n_verts_cumsum
                )

            centers_blocks.append(centers)
            offsets_blocks.append(offsets)
            triangles_blocks.append(triangles)
            n_verts_cumsum += centers.shape[0]

        return (
            np.concatenate(centers_blocks),
            np.concatenate(offsets_blocks),
            np.concatenate(triangles_blocks),
        )

    def shapes_in_box(
        self, corners: np.ndarray[tuple[Literal[2], Literal[2]]]
    ) -> list[int]:
        """Determines which shapes, if any, are inside an axis-aligned box.

        Looks only at displayed shapes

        Parameters
        ----------
        corners : np.ndarray
            2x2 array of two corners that will be used to create an
            axis-aligned box.

        Returns
        -------
        shapes : list of ints
            List of shapes that are inside the box.
        """
        shape_mins, shape_maxs = self._bounding_boxes
        if shape_mins.shape[0] == 0:
            return []

        selection_min = np.min(corners, axis=0)
        selection_max = np.max(corners, axis=0)

        # If the box encompasses all shapes, just get them directly
        layer_min = np.min(shape_mins, axis=0)
        layer_max = np.max(shape_maxs, axis=0)
        if np.all(selection_min <= layer_min) and np.all(
            selection_max >= layer_max
        ):
            return self._visible_shapes_indices.tolist()

        # Get shapes with bounding boxes intersecting the selection box
        intersects_mask = np.all(shape_maxs >= selection_min, axis=1) & np.all(
            shape_mins <= selection_max, axis=1
        )

        intersecting_indices = self._visible_shapes_indices[intersects_mask]

        if intersecting_indices.size == 0:
            return []

        shape_mins = shape_mins[intersects_mask]
        shape_maxs = shape_maxs[intersects_mask]

        shapes_full_in_mask = np.all(
            shape_maxs <= selection_max, axis=1
        ) & np.all(shape_mins >= selection_min, axis=1)

        return [
            num
            for num, full_in in zip(
                intersecting_indices, shapes_full_in_mask, strict=True
            )
            if full_in
            or triangles_intersect_box(
                self._mesh.vertices[
                    self._mesh.triangles[self._mesh_triangles_slice(num)]
                ],
                corners,
            ).any()
        ]

    @cached_property
    def _visible_shapes(self) -> list[tuple[int, Shape]]:
        slice_key = self.slice_key
        if len(slice_key):
            return [
                (i, s)
                for i, s in enumerate(self.shapes)
                if (
                    np.all(s.slice_key[0] <= slice_key)
                    and np.all(slice_key <= s.slice_key[1])
                )
            ]
        return list(enumerate(self.shapes))

    @cached_property
    def _bounding_boxes(
        self,
    ) -> tuple[
        np.ndarray[tuple[int, Literal[2, 3]]],
        np.ndarray[tuple[int, Literal[2, 3]]],
    ]:
        data = np.array([s[1].bounding_box for s in self._visible_shapes])
        if data.size == 0:
            return np.empty((0, self.ndisplay)), np.empty((0, self.ndisplay))  # type: ignore[return-value]
        return data[:, 0], data[:, 1]

    @cached_property
    def _visible_shapes_indices(
        self,
    ) -> np.ndarray[tuple[int], np.dtype[IndexDtype]]:
        return np.array([s[0] for s in self._visible_shapes])

    def inside(self, coord):
        """Determines if any shape at given coord by looking inside triangle
        meshes. Looks only at displayed shapes

        Parameters
        ----------
        coord : sequence of float
            Image coordinates to check if any shapes are at.

        Returns
        -------
        shape : int | None
            Index of shape if any that is at the coordinates. Returns `None`
            if no shape is found.
        """
        if not self.shapes:
            return None
        bounding_boxes = self._bounding_boxes
        in_bbox = np.all(
            (bounding_boxes[0] <= coord) * (bounding_boxes[1] >= coord),
            axis=1,
        )
        inside_indices = np.flatnonzero(in_bbox)
        if inside_indices.size == 0:
            return None
        try:
            z_index = [
                self._visible_shapes[i][1].z_index for i in inside_indices
            ]
            pos = np.argsort(z_index)
            return self._visible_shapes[
                next(
                    inside_indices[p]
                    for p in pos[::-1]
                    if np.any(
                        inside_triangles(
                            self._visible_shapes[inside_indices[p]][
                                1
                            ]._all_triangles()
                            - coord
                        )
                    )
                )
            ][0]
        except StopIteration:
            return None

    def _inside_3d(self, ray_position: np.ndarray, ray_direction: np.ndarray):
        """Determines if any shape is intersected by a ray by looking inside triangle
        meshes. Looks only at displayed shapes.

        Parameters
        ----------
        ray_position : np.ndarray
            (3,) array containing the location that was clicked. This
            should be in the same coordinate system as the vertices.
        ray_direction : np.ndarray
            (3,) array describing the direction camera is pointing in
            the scene. This should be in the same coordinate system as
            the vertices.

        Returns
        -------
        shape : int | None
            Index of shape if any that is at the coordinates. Returns `None`
            if no shape is found.
        intersection_point : Optional[np.ndarray]
            The point where the ray intersects the mesh face. If there was
            no intersection, returns None.
        """
        triangles = self._mesh.vertices[self._mesh.displayed_triangles]
        inside = line_in_triangles_3d(
            line_point=ray_position,
            line_direction=ray_direction,
            triangles=triangles,
        )
        # intersected_shapes = self._mesh.displayed_triangles_index[inside, 0]
        if not np.any(inside):
            return None, None

        intersection_points = self._triangle_intersection(
            triangle_indices=inside,
            ray_position=ray_position,
            ray_direction=ray_direction,
        )
        start_to_intersection = intersection_points - ray_position
        distances = np.linalg.norm(start_to_intersection, axis=1)
        closest_shape_index = np.argmin(distances)
        shape = self._mesh.displayed_triangles_to_shape_index[inside][
            closest_shape_index
        ]
        intersection = intersection_points[closest_shape_index]
        return shape, intersection

    def _triangle_intersection(
        self,
        triangle_indices: np.ndarray,
        ray_position: np.ndarray,
        ray_direction: np.ndarray,
    ):
        """Find the intersection of a ray with specified triangles.

        Parameters
        ----------
        triangle_indices : np.ndarray
            (n,) array of shape indices to find the intersection with the ray. The indices should
            correspond with self._mesh.displayed_triangles.
        ray_position : np.ndarray
            (3,) array with the coordinate of the starting point of the ray in layer coordinates.
            Only provide the 3 displayed dimensions.
        ray_direction : np.ndarray
            (3,) array of the normal direction of the ray in layer coordinates.
            Only provide the 3 displayed dimensions.

        Returns
        -------
        intersection_points : np.ndarray
            (n x 3) array of the intersection of the ray with each of the specified shapes in layer coordinates.
            Only the 3 displayed dimensions are provided.
        """
        triangles = self._mesh.vertices[self._mesh.displayed_triangles]
        intersected_triangles = triangles[triangle_indices]
        intersection_points = intersect_line_with_triangles(
            line_point=ray_position,
            line_direction=ray_direction,
            triangles=intersected_triangles,
        )
        return intersection_points

    def to_masks(self, mask_shape=None, zoom_factor=1, offset=(0, 0)):
        """Returns N binary masks, one for each shape, embedded in an array of
        shape `mask_shape`.

        Parameters
        ----------
        mask_shape : np.ndarray | tuple | None
            2-tuple defining shape of mask to be generated. If non specified,
            takes the max of all the vertices
        zoom_factor : float
            Premultiplier applied to coordinates before generating mask. Used
            for generating as downsampled mask.
        offset : 2-tuple
            Offset subtracted from coordinates before multiplying by the
            zoom_factor. Used for putting negative coordinates into the mask.

        Returns
        -------
        masks : (N, M, P) np.ndarray
            Array where there is one binary mask of shape MxP for each of
            N shapes
        """
        if mask_shape is None:
            mask_shape = self.displayed_vertices.max(axis=0).astype('int')

        masks = np.array(
            [
                s.to_mask(mask_shape, zoom_factor=zoom_factor, offset=offset)
                for s in self.shapes
            ]
        )

        return masks

    def to_labels(self, labels_shape=None, zoom_factor=1, offset=(0, 0)):
        """Returns a integer labels image, where each shape is embedded in an
        array of shape labels_shape with the value of the index + 1
        corresponding to it, and 0 for background. For overlapping shapes
        z-ordering will be respected.

        Parameters
        ----------
        labels_shape : np.ndarray | tuple | None
            2-tuple defining shape of labels image to be generated. If non
            specified, takes the max of all the vertices
        zoom_factor : float
            Premultiplier applied to coordinates before generating mask. Used
            for generating as downsampled mask.
        offset : 2-tuple
            Offset subtracted from coordinates before multiplying by the
            zoom_factor. Used for putting negative coordinates into the mask.

        Returns
        -------
        labels : np.ndarray
            MxP integer array where each value is either 0 for background or an
            integer up to N for points inside the corresponding shape.
        """
        if labels_shape is None:
            labels_shape = self.displayed_vertices.max(axis=0).astype(int)

        labels = np.zeros(labels_shape, dtype=int)

        for ind in self._z_order[::-1]:
            mask = self.shapes[ind].to_mask(
                labels_shape, zoom_factor=zoom_factor, offset=offset
            )
            labels[mask] = ind + 1

        return labels

    def to_colors(
        self, colors_shape=None, zoom_factor=1, offset=(0, 0), max_shapes=None
    ):
        """Rasterize shapes to an RGBA image array.

        Each shape is embedded in an array of shape `colors_shape` with the
        RGBA value of the shape, and 0 for background. For overlapping shapes
        z-ordering will be respected.

        Parameters
        ----------
        colors_shape : np.ndarray | tuple | None
            2-tuple defining shape of colors image to be generated. If non
            specified, takes the max of all the vertiecs
        zoom_factor : float
            Premultiplier applied to coordinates before generating mask. Used
            for generating as downsampled mask.
        offset : 2-tuple
            Offset subtracted from coordinates before multiplying by the
            zoom_factor. Used for putting negative coordinates into the mask.
        max_shapes : None | int
            If provided, this is the maximum number of shapes that will be rasterized.
            If the number of shapes in view exceeds max_shapes, max_shapes shapes
            will be randomly selected from the in view shapes. If set to None, no
            maximum is applied. The default value is None.

        Returns
        -------
        colors : (N, M, 4) array
            rgba array where each value is either 0 for background or the rgba
            value of the shape for points inside the corresponding shape.
        """
        if colors_shape is None:
            colors_shape = self.displayed_vertices.max(axis=0).astype(int)

        colors = np.zeros((*colors_shape, 4), dtype=float)
        colors[..., 3] = 1

        z_order = self._z_order
        shapes_in_view = np.argwhere(self._displayed)
        z_order_in_view_mask = np.isin(z_order, shapes_in_view)
        z_order_in_view = z_order[z_order_in_view_mask]

        # If there are too many shapes to render responsively, just render
        # the top max_shapes shapes
        if max_shapes is not None and len(z_order_in_view) > max_shapes:
            z_order_in_view = z_order_in_view[-max_shapes:]

        for ind in z_order_in_view:
            mask = self.shapes[ind].to_mask(
                colors_shape, zoom_factor=zoom_factor, offset=offset
            )
            if type(self.shapes[ind]) in [Path, Line]:
                col = self._edge_color[ind]
            else:
                col = self._face_color[ind]
            colors[mask, :] = col

        return colors

    def _clear_cache(self):
        self.__dict__.pop('_bounding_boxes', None)
        self.__dict__.pop('_visible_shapes', None)
        self.__dict__.pop('_visible_shapes_indices', None)



"""
Layers
======

Display multiple image layers using the ``add_image`` API and then reorder them
using the layers swap method and remove one

.. tags:: visualization-basic
"""

import numpy as np
from skimage import data
from skimage.color import rgb2gray

import napari

# create the viewer with several image layers
viewer = napari.Viewer()
layer = viewer.add_image(rgb2gray(data.astronaut()), name='astronaut')
viewer.add_image(data.camera(), name='photographer')
viewer.add_image(data.coins(), name='coins')
viewer.add_image(data.moon(), name='moon')
viewer.add_image(np.random.random((512, 512)), name='random')
viewer.add_image(data.binary_blobs(length=512, volume_fraction=0.2, n_dim=2), name='blobs')
viewer.grid.enabled = True

if __name__ == '__main__':
    napari.run()



"""
Fourier transform playground
============================

Generate an image by adding arbitrary 2D sine waves and observe
how the fourier transform's real and imaginary components are affected
by the changes. Threading is used to smoothly animate the waves.

.. tags:: interactivity, gui
"""


from time import sleep, time

import numpy as np
from magicgui import magic_factory
from qtpy.QtWidgets import QApplication
from scipy.fft import fft2, fftshift

import napari
from napari.qt.threading import thread_worker

IMAGE_SIZE = 100
FPS = 20
# when running this example in sphinx-gallery, we need to tell the
# thread when the example is finished running to stop updates
FINISHED = False

# meshgrid used to calculate the 2D sine waves
x = np.arange(IMAGE_SIZE) - IMAGE_SIZE / 2
X, Y = np.meshgrid(x, x)

def wait_for_layers(viewer, layer_names, *, interval=0.1):
    """Wait for any thread creating layers to finish.

    If we expect a different thread to be adding layers to
    `viewer`, we can wait for them to be added based on
    their expected names.
    """
    for name in layer_names:
        while name not in viewer.layers:
            sleep(interval)
            QApplication.processEvents()

    QApplication.processEvents()


def wave_2d(frequency, angle, phase_shift):
    """Generate a 2D sine wave based on angle, frequency and phase shift."""
    angle = np.deg2rad(angle)
    phase_shift = np.deg2rad(phase_shift)
    wave = 2 * np.pi * (X * np.cos(angle) + Y * np.sin(angle)) * frequency
    return np.sin(wave + phase_shift)


# set up viewer with grid-mode enabled
viewer = napari.Viewer()
viewer.grid.enabled = True


def update_layer(name, data, **kwargs):
    """Update a layer in the viewer with new data.

    If data is None, then the layer is removed.
    If the layer is not present, it's added to the viewer.
    """
    if FINISHED:
        return # don't update the viewer after the example has finished

    if data is None:
        if name in viewer.layers:
            viewer.layers.pop(name)
        viewer.reset_view()
    elif name not in viewer.layers:
        viewer.add_image(data, name=name, interpolation2d='spline36', **kwargs)
        viewer.reset_view()
    else:
        viewer.layers[name].data = data


def combine_and_set_data(wave_args):
    """Merge 2D waves, calculate the FT and update the viewer.

    The wave phases are offset by the current time multiplied by an
    arbitrary speed value; this generates an animated wave if called repeatedly.
    """
    if not wave_args:
        # this happens on yielding from the thread, no need to update anything
        return

    t = time()
    waves = {
        wave_id: wave_2d(frequency, angle, phase_shift + t * speed * 100) if frequency else None
        for wave_id, (frequency, angle, phase_shift, speed) in wave_args.items()
    }

    to_add = [d for d in waves.values() if d is not None]
    if to_add:
        mean = np.mean(to_add, axis=0)
        ft = fftshift(fft2(mean))
        power_spectrum = abs(ft)
        phase = np.angle(ft) * power_spectrum
        power_spectrum = np.log10(power_spectrum + 10)
    else:
        mean = power_spectrum = phase = None

    # for visualisation, it's clearer to use:
    # phase * ps instead of phase
    # and log10(ps + 1) instead of ps
    update_layer('phase * power_spectrum', phase, colormap=('blue', 'black', 'red'))
    update_layer('log10(power_spectrum + 1)', power_spectrum)
    update_layer('mean', mean)

    for name, data in waves.items():
        update_layer(f'wave {name}', data)


@thread_worker(connect={'yielded': combine_and_set_data})
def update_viewer():
    # keep track of each wave in a dictionary by id, this way we can modify/remove
    # existing waves or add new ones
    wave_args = {}
    new_params = None
    while True:
        sleep(1 / FPS)
        # see https://napari.org/stable/guides/threading.html#full-two-way-communication
        # this receives new_params from thread.send() and yields {} for the `yielded` callback
        new_params = yield wave_args
        if new_params is not None:
            # note that these come from thread.send() in moving_wave()!
            wave_id, *args = new_params
            wave_args[wave_id] = args
        yield wave_args


# start the thread responsible for updating the viewer
thread = update_viewer()


@magic_factory(
    auto_call=True,
    frequency={'widget_type': 'FloatSlider', 'min': 0, 'max': 1, 'step': 0.01},
    angle={'widget_type': 'Slider', 'min': 0, 'max': 180},
    phase_shift={'widget_type': 'Slider', 'min': 0, 'max': 180},
    speed={'widget_type': 'FloatSlider', 'min': -10, 'max': 10, 'step': 0.1},
)
def moving_wave(
    wave_id: int = 0,
    frequency: float = 0.2,
    angle: int = 0,
    phase_shift: int = 0,
    speed: float = 1,
    run=True,
):
    """Send new parameters to the listening thread to update the 2D waves.

    The `run` checkbox can be disabled to stop sending values to the thread
    while changing parameters.
    """
    if run:
        thread.send((wave_id, frequency, angle, phase_shift, speed))


wdg = moving_wave()

# add the widget to the window and run it once
viewer.window.add_dock_widget(wdg, area='bottom')
wdg()

# wait for the layers to be added before running the viewer
wait_for_layers(viewer, ['wave 0'])

if __name__ == '__main__':
    napari.run()

thread.quit()
FINISHED = True




import warnings

import napari
from napari._qt.widgets.qt_viewer_buttons import QtViewerPushButton


def raise_():
    x = 1
    y = 'a string'
    import something_that_does_not_exist
    return something_that_does_not_exist.fun(x, y)


def warn_():
    warnings.warn('warning!')


viewer = napari.Viewer()
layer_buttons = viewer.window._qt_viewer.layerButtons
err_btn = QtViewerPushButton('warning', 'new Error', raise_)
warn_btn = QtViewerPushButton('warning', 'new Warn', warn_)
layer_buttons.layout().insertWidget(3, warn_btn)
layer_buttons.layout().insertWidget(3, err_btn)

if __name__ == '__main__':
    napari.run()



import warnings

import numpy as np
from magicgui import magicgui
from vispy.scene.visuals import Ellipse

import napari
from napari._vispy.overlays.base import ViewerOverlayMixin, VispyCanvasOverlay
from napari._vispy.utils.visual import overlay_to_visual
from napari.components._viewer_constants import CanvasPosition
from napari.components.overlays import CanvasOverlay
from napari.utils.color import ColorValue


# the overlay model should inherit from either CanvasOverlay or SceneOverlay
# depending on if it needs to live in "screen space" or "scene space"
# (i.e: if it should be affected by camera, dims, ndisplay, ...)
class DotOverlay(CanvasOverlay):
    """
    Example overlay using a colored dot to show some state
    """
    color: ColorValue = (0, 1, 0, 1)
    size: int = 10


# the vispy overlay class should handle connecting the model to the vispy visual
# we use the ViewerOverlayMixin because this overlay is attached to the viewer,
# and not a specific layer
class VispyDotOverlay(ViewerOverlayMixin, VispyCanvasOverlay):
    # all arguments are keyword-only. viewer, overlay and parent should always be present.
    def __init__(self, *, viewer, overlay, parent=None):
        # the node argument for the base class is the vispy visual
        # note that the center is (0, 0), cause we handle the shift with transforms
        super().__init__(
            node=Ellipse(center=(0, 0)),
            viewer=viewer,
            overlay=overlay,
            parent=parent,
        )

        # we also need to connect events from the model to callbacks that update the visual
        self.overlay.events.color.connect(self._on_color_change)
        self.overlay.events.size.connect(self._on_size_change)
        # no need to connect position, since that's in the base classes of CanvasOverlay

        # at the end of the init of subclasses of VispyBaseOverlay we always
        # need to call reset to initialize properly
        self.reset()

    def _on_color_change(self, event=None):
        self.node.color = self.overlay.color

    def _on_position_change(self, event=None):
        # we can overload the position changing to account for the size, so that the dot
        # always sticks to the edge; there are `offset` attributes specifically for this
        self.x_offset = self.y_offset = self.overlay.size / 2
        super()._on_position_change()

    def _on_size_change(self, event=None):
        self.node.radius = self.overlay.size / 2
        self.x_size = self.overlay.size
        self.y_size = self.overlay.size
        # trigger position update since the radius changed
        self._on_position_change()

    # we should always add all the new callbacks to the reset() method
    def reset(self):
        super().reset()
        self._on_color_change()
        self._on_size_change()


# for napari to know how to use this overlay, we need to add it to the overlay_to_visual dict
# this will ideally be exposed at some point
overlay_to_visual[DotOverlay] = VispyDotOverlay

viewer = napari.Viewer()
# we also need to add at least a layer to see any overlay,
# since the canvas is otherwise covered by the welcome widget
viewer.add_image(np.random.rand(10, 10))

# note that we're accessing private attributes externally, which triggers a bunch of warnings.
# suppress them for the purpose of this example
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    viewer._overlays['dot'] = DotOverlay(visible=True)


# let's make a simple widget to control the overlay
@magicgui(
    auto_call=True,
    color={'choices': ['red', 'blue', 'green', 'magenta']},
)
def control_dot(viewer: napari.Viewer, color='red', position: CanvasPosition = 'top_left'):
    dot = viewer._overlays['dot']
    dot.color = color
    dot.position = position


viewer.window.add_dock_widget(control_dot)


# and let's also add a mouse callback to do something when dragging the mouse
def change_size(viewer, event):
    pos = np.array(event.pos)
    size = viewer._overlays['dot'].size

    # use event.handled to tell vispy to not drag the canvas
    event.handled = True

    yield

    while event.type == 'mouse_move':
        new_pos = event.pos
        drag = new_pos[0] - pos[0]

        viewer._overlays['dot'].size = size + drag
        yield


viewer.mouse_drag_callbacks.append(change_size)

if __name__ == '__main__':
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        control_dot()
        napari.run()



from qtpy.QtCore import Qt
from qtpy.QtWidgets import QComboBox, QWidget

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Tracks
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtColorPropertiesComboBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer color properties
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Tracks
        An instance of a napari Tracks layer.

    Attributes
    ----------
    color_by_combobox : qtpy.QtWidgets.QComboBox
        ComboBox controlling current color property of the layer.
    color_by_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the color property chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Tracks) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.color_by.connect(self._on_color_by_change)
        self._layer.events.properties.connect(self._on_properties_change)

        # Setup widgets
        # combo box for track coloring, we can get these from the properties
        # keys
        self.color_by_combobox = QComboBox()
        self.color_by_combobox.addItems(self._layer.properties_to_color_by)
        connect_setattr(
            self.color_by_combobox.currentTextChanged, self._layer, 'color_by'
        )

        self.color_by_combobox_label = QtWrappedLabel(trans._('color by:'))

        self._on_color_by_change()

    def _on_color_by_change(self) -> None:
        """Receive layer model color_by change event and update combobox."""
        with qt_signals_blocked(self.color_by_combobox):
            color_by = self._layer.color_by

            idx = self.color_by_combobox.findText(
                color_by, Qt.MatchFlag.MatchFixedString
            )
            self.color_by_combobox.setCurrentIndex(idx)

    def _on_properties_change(self) -> None:
        """Change the properties that can be used to color the tracks."""
        with qt_signals_blocked(self.color_by_combobox):
            self.color_by_combobox.clear()
            self.color_by_combobox.addItems(self._layer.properties_to_color_by)
        self._on_color_by_change()

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.color_by_combobox_label, self.color_by_combobox)]



import logging

from vispy.scene.visuals import Markers as BaseMarkers

logger = logging.getLogger(__name__)


class Markers(BaseMarkers):
    def _compute_bounds(self, axis, view):
        # needed for entering 3D rendering mode when a points
        # layer is invisible and the self._data property is None
        if self._data is None:
            return None
        pos = self._data['a_position']
        if pos is None or pos.size == 0:
            return None
        if pos.shape[1] > axis:
            return (pos[:, axis].min(), pos[:, axis].max())

        return (0, 0)



# See "Writing benchmarks" in the asv docs for more information.
# https://asv.readthedocs.io/en/latest/writing_benchmarks.html
# or the napari documentation on benchmarking
# https://github.com/napari/napari/blob/main/docs/BENCHMARKS.md
import time

import numpy as np
import zarr
from qtpy.QtWidgets import QApplication

import napari
from napari.layers import Image

from .utils import Skip

SAMPLE_PARAMS = {
    'skin_data': {
        # napari-bio-sample-data
        'shape': (1280, 960, 3),
        'chunk_shape': (512, 512, 3),
        'dtype': 'uint8',
    },
    'jrc_hela-2 (scale 3)': {
        # s3://janelia-cosem-datasets/jrc_hela-2/jrc_hela-2.n5
        'shape': (796, 200, 1500),
        'dtype': 'uint16',
        'chunk_shape': (64, 64, 64),
    },
}


def get_image_params():
    # chunksizes = [(64,64,64), (256,256,256), (512,512,512)]
    latencies = [0.05 * i for i in range(3)]
    datanames = SAMPLE_PARAMS.keys()
    params = (latencies, datanames)

    return params


class SlowMemoryStore(zarr.storage.MemoryStore):
    def __init__(self, load_delay, *args, **kwargs) -> None:
        self.load_delay = load_delay
        super().__init__(*args, **kwargs)

    def __getitem__(self, item: str):
        time.sleep(self.load_delay)
        return super().__getitem__(item)


class AsyncImage2DSuite:
    params = get_image_params()
    timeout = 300
    skip_params = Skip(if_in_pr=lambda latency, dataname: latency > 0)

    def setup(self, latency, dataname):
        shape = SAMPLE_PARAMS[dataname]['shape']
        chunk_shape = SAMPLE_PARAMS[dataname]['chunk_shape']
        dtype = SAMPLE_PARAMS[dataname]['dtype']

        store = SlowMemoryStore(load_delay=latency)
        self.data = zarr.zeros(
            shape,
            chunks=chunk_shape,
            dtype=dtype,
            store=store,
        )

        self.layer = Image(self.data)

    def time_create_layer(self, *args):
        """Time to create an image layer."""
        Image(self.data)

    def time_set_view_slice(self, *args):
        """Time to set view slice."""
        if hasattr(self.layer, '_slicing_state'):
            self.layer._slicing_state._set_view_slice()
        else:
            # before https://github.com/napari/napari/pull/8254
            self.layer._set_view_slice()

    def time_refresh(self, *args):
        """Time to refresh view."""
        self.layer.refresh()


def _skip_3d_rgb(_latency, dataname):
    shape = SAMPLE_PARAMS[dataname]['shape']
    return len(shape) == 3 and shape[2] == 3


class QtViewerAsyncImage2DSuite:
    params = get_image_params()
    skip_params = Skip(
        always=_skip_3d_rgb, if_in_pr=lambda latency, dataname: latency > 0
    )
    timeout = 300

    def setup(self, latency, dataname):
        shape = SAMPLE_PARAMS[dataname]['shape']
        chunk_shape = SAMPLE_PARAMS[dataname]['chunk_shape']
        dtype = SAMPLE_PARAMS[dataname]['dtype']

        store = SlowMemoryStore(load_delay=latency)
        _ = QApplication.instance() or QApplication([])
        self.data = zarr.zeros(
            shape,
            chunks=chunk_shape,
            dtype=dtype,
            store=store,
        )

        self.viewer = napari.Viewer()
        self.viewer.add_image(self.data)

    def time_z_scroll(self, *args):
        layers_to_scroll = 4
        for z in range(layers_to_scroll):
            z = z * (self.data.shape[2] // layers_to_scroll)
            self.viewer.dims.set_current_step(0, z)

    def teardown(self, *args):
        if self.viewer is not None:
            self.viewer.window.close()


class QtViewerAsyncPointsSuite:
    n_points = [2**i for i in range(12, 18)]
    params = n_points
    skip_params = Skip(if_in_pr=lambda n_points: n_points > 2**12)

    def setup(self, n_points):
        _ = QApplication.instance() or QApplication([])

        np.random.seed(0)
        self.viewer = napari.Viewer()
        # Fake image layer to set bounds. Is this really needed?
        self.empty_image = np.zeros((512, 512, 512), dtype='uint8')
        self.viewer.add_image(self.empty_image)
        self.point_data = np.random.randint(512, size=(n_points, 3))
        self.viewer.add_points(self.point_data)
        self.app = QApplication.instance() or QApplication([])

    def time_z_scroll(self, *args):
        for z in range(self.empty_image.shape[0]):
            self.viewer.dims.set_current_step(0, z)
            self.app.processEvents()

    def teardown(self, *args):
        self.viewer.window.close()


class QtViewerAsyncPointsAndImage2DSuite:
    n_points = [2**i for i in range(12, 18, 2)]
    chunksize = [256, 512, 1024]
    latency = [0.05 * i for i in range(3)]
    params = (n_points, latency, chunksize)
    timeout = 600

    skip_params = Skip(
        if_in_pr=lambda n_points, latency, chunksize: n_points > 2**14
        or chunksize > 512
        or latency > 0,
    )

    def setup(self, n_points, latency, chunksize):
        store = SlowMemoryStore(load_delay=latency)
        _ = QApplication.instance() or QApplication([])

        np.random.seed(0)

        self.image_data = zarr.zeros(
            (64, 2048, 2048),
            chunks=(1, chunksize, chunksize),
            dtype='uint8',
            store=store,
        )

        self.viewer = napari.Viewer()
        self.viewer.add_image(self.image_data)
        self.point_data = np.random.randint(512, size=(n_points, 3))
        self.viewer.add_points(self.point_data)
        self.app = QApplication.instance() or QApplication([])

    def time_z_scroll(self, *args):
        for z in range(self.image_data.shape[0]):
            self.viewer.dims.set_current_step(0, z)
            self.app.processEvents()

    def teardown(self, *args):
        self.viewer.window.close()


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



"""
Embed IPython
=============

Start napari and land directly in an embedded ipython console with qt event loop.

A similar effect can be achieved more simply with `viewer.update_console(locals())`,
such as shown in https://github.com/napari/napari/blob/main/examples/update_console.py.

However, differently from `update_console`, this will start an independent
ipython console which can outlive the viewer.

.. tags:: gui
"""

from IPython.terminal.embed import InteractiveShellEmbed

import napari

# any code
text = 'some text'

# initialize viewer
viewer = napari.Viewer()

# embed ipython and run the magic command to use the qt event loop
sh = InteractiveShellEmbed()
sh.enable_gui('qt')  # equivalent to using the '%gui qt' magic
sh()  # open the embedded shell

# From there, you can access the script's scope, such as the variables `text` and `viewer`



import subprocess
import sys


class ImportTimeSuite:
    def time_import(self):
        cmd = [sys.executable, '-c', 'import napari']
        subprocess.run(cmd, stderr=subprocess.PIPE)


if __name__ == '__main__':
    from utils import run_benchmark

    run_benchmark()



"""
Dask nD image
=============

Display a dask array

.. tags:: visualization-nD
"""

try:
    from dask import array as da
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        """This example uses a dask array but dask is not
    installed. To install try 'pip install dask'."""
    ) from None

import numpy as np
from skimage import data

import napari

blobs = da.stack(
    [
        data.binary_blobs(
            length=128, blob_size_fraction=0.05, n_dim=3, volume_fraction=f
        )
        for f in np.linspace(0.05, 0.5, 10)
    ],
    axis=0,
)
viewer = napari.Viewer()
layer = viewer.add_image(blobs.astype(float))

if __name__ == '__main__':
    napari.run()



"""
Add shapes
==========

Display one shapes layer ontop of one image layer using the ``add_shapes`` and
``add_image`` APIs. When the window is closed it will print the coordinates of
your shapes.

.. tags:: visualization-basic
"""

import numpy as np
from skimage import data

import napari

# add the image
viewer = napari.Viewer()
layer = viewer.add_image(data.camera(), name='photographer')

# create a list of polygons
polygons = [
    np.array([[11, 13], [111, 113], [22, 246]]),
    np.array(
        [
            [505, 60],
            [402, 71],
            [383, 42],
            [251, 95],
            [212, 59],
            [131, 137],
            [126, 187],
            [191, 204],
            [171, 248],
            [211, 260],
            [273, 243],
            [264, 225],
            [430, 173],
            [512, 160],
        ]
    ),
    np.array(
        [
            [310, 382],
            [229, 381],
            [209, 401],
            [221, 411],
            [258, 411],
            [300, 412],
            [306, 435],
            [268, 434],
            [265, 454],
            [298, 461],
            [307, 461],
            [307, 507],
            [349, 510],
            [352, 369],
            [330, 366],
        ]
    ),
]

# add polygons
layer = viewer.add_shapes(
    polygons,
    shape_type='polygon',
    edge_width=1,
    edge_color='coral',
    face_color='royalblue',
    name='shapes',
)

# shapes of each type can also be added via their respective add_<shape_type> method
# e.g. for the polygons above:

# layer = viewer.add_shapes(name='shapes') # create empty layer
# layer.add_polygons(
#     polygons,
#     edge_width=1,
#     edge_color='coral',
#     face_color='royalblue',
# )


# change some attributes of the layer
layer.selected_data = set(range(layer.nshapes))
layer.current_edge_width = 5
layer.selected_data = set()

# add an ellipse to the layer
ellipse = np.array([[59, 222], [110, 289], [170, 243], [119, 176]])
layer.add(
    ellipse,
    shape_type='ellipse',
    edge_width=5,
    edge_color='coral',
    face_color='purple',
)

# To save layers to svg:
# viewer.layers.save('viewer.svg', plugin='svg')

if __name__ == '__main__':
    napari.run()



"""
nD shapes with text
===================

.. tags:: visualization-nD
"""
from skimage import data

import napari

blobs = data.binary_blobs(
    length=100, blob_size_fraction=0.05, n_dim=3, volume_fraction=0.03
).astype(float)

viewer = napari.Viewer(ndisplay=3)
layer = viewer.add_image(blobs.astype(float))

n = 50
shape = [[[n, 40, 40], [n, 40, 60], [n + 20, 60, 60], [n + 20, 60, 40]]]

features = {'z_index': [n]}
text = {'string': 'z_index', 'color': 'green', 'anchor': 'upper_left'}

shapes_layer = viewer.add_shapes(
    shape,
    edge_color=[0, 1, 0, 1],
    face_color='transparent',
    features=features,
    text=text,
)

if __name__ == '__main__':
    napari.run()



import csv

import numpy as np
import pooch

import napari

"""
This data comes from the Neurocyto Lab's description of the ThunderSTORM format.
This file format is used to represent single molecule localizations.

With respect to the napari async slicing work, this dataset is small enough that it performs well in synchronous mode.

If someone is interested, then you can use the uncertainty_xy attribute from the STORM data to change the point size.

More information is available here: http://www.neurocytolab.org/tscolumns/
"""

storm_path = pooch.retrieve(
    url='http://www.neurocytolab.org/wp-content/uploads/2018/06/ThunderSTORM_TS3D.csv',
    known_hash='665a28b2fad69dbfd902e4945df04667f876d33a91167614c280065212041a29',
    progressbar=True
)

with open(storm_path) as csvfile:
    data = list(csv.reader(csvfile))

data = np.array(data[1:]).astype(float)
data = data[:, 1:4]
viewer = napari.Viewer()
layer = viewer.add_points(data, size = 50)
napari.run()



"""
Matplotlib plot
===============

.. tags:: gui
"""

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_qt5agg import FigureCanvas

import napari

# create image
x = np.linspace(0, 5, 256)
y = np.linspace(0, 5, 256)[:, np.newaxis]
img = np.sin(x) ** 10 + np.cos(10 + y * x) * np.cos(x)

# add it to the viewer
viewer = napari.Viewer()
layer = viewer.add_image(img, colormap='viridis')

# create mpl figure with subplots
mpl_fig = plt.figure()
ax = mpl_fig.add_subplot(111)
(line,) = ax.plot(layer.data[123])  # linescan through the middle of the image

# add the figure to the viewer as a FigureCanvas widget
viewer.window.add_dock_widget(FigureCanvas(mpl_fig))


# connect a callback that updates the line plot when
# the user clicks on the image
@layer.mouse_drag_callbacks.append
def profile_lines_drag(layer, event):
    try:
        line.set_ydata(layer.data[int(event.position[0])])
        line.figure.canvas.draw()
    except IndexError:
        pass


if __name__ == '__main__':
    napari.run()



"""Example of using low-level `QtListView` with SelectableEventedList

:class:`napari.utils.events.SelectableEventedList` is a mutable sequence that
emits events when modified.  It also has a selection model (tracking which
items are selected).

:class:`napari._qt.containers.QtListView` adapts the `EventedList` to the
QAbstractItemModel/QAbstractItemView interface used by the QtFramework.  This
allows you to create an interactive GUI view onto a python model that stays
up to date, and can modify the python object... while maintining the python
object as the single "source of truth".
"""
import napari
from napari._qt.containers import QtListView
from napari.qt import get_qapp
from napari.utils.events import SelectableEventedList

get_qapp()


class MyObject:
    """generic object."""

    def __init__(self, name) -> None:
        self.name = name

    def __str__(self):
        return self.name


# create our evented list
root = SelectableEventedList([MyObject(x) for x in 'abcdefg'])
# create Qt view onto the list
view = QtListView(root)
# show the view
view.show()


# spy on events
root.events.reordered.connect(lambda e: print('reordered to: ', e.value))
root.selection.events.changed.connect(
    lambda e: print(
        f'selection changed.  added: {e.added}, removed: {e.removed}'
    )
)
root.selection.events._current.connect(
    lambda e: print(f'current item changed to: {e.value}')
)

if __name__ == '__main__':
    napari.run()



"""
Add 3D image
============

Display a 3D image layer using the :meth:`add_image` API.

.. tags:: visualization-nD, layers
"""

from skimage import data

import napari

blobs = data.binary_blobs(length=64, volume_fraction=0.1, n_dim=3).astype(
    float
)
viewer = napari.Viewer(ndisplay=3)
# add the volume
viewer.add_image(blobs, scale=[3, 1, 1])

if __name__ == '__main__':
    napari.run()



from app_model import Action
from app_model.types import StandardKeyBinding, SubmenuItem, ToggleRule

from napari._app_model.actions._toggle_action import ViewerModelToggleAction
from napari._app_model.constants import MenuGroup, MenuId
from napari.components import ViewerModel
from napari.settings import get_settings
from napari.utils.translations import trans

VIEW_SUBMENUS = [
    (
        MenuId.MENUBAR_VIEW,
        SubmenuItem(submenu=MenuId.VIEW_AXES, title=trans._('Axes')),
    ),
    (
        MenuId.MENUBAR_VIEW,
        SubmenuItem(submenu=MenuId.VIEW_SCALEBAR, title=trans._('Scale Bar')),
    ),
]


toggle_action_details = [
    (
        'napari.window.view.toggle_viewer_axes',
        trans._('Axes Visible'),
        'axes',
        'visible',
    ),
    (
        'napari.window.view.toggle_viewer_axes_colored',
        trans._('Axes Colored'),
        'axes',
        'colored',
    ),
    (
        'napari.window.view.toggle_viewer_axes_labels',
        trans._('Axes Labels'),
        'axes',
        'labels',
    ),
    (
        'napari.window.view.toggle_viewer_axes_dashed',
        trans._('Axes Dashed'),
        'axes',
        'dashed',
    ),
    (
        'napari.window.view.toggle_viewer_axes_arrows',
        trans._('Axes Arrows'),
        'axes',
        'arrows',
    ),
    (
        'napari.window.view.toggle_viewer_scale_bar',
        trans._('Scale Bar Visible'),
        'scale_bar',
        'visible',
    ),
    (
        'napari.window.view.toggle_viewer_scale_bar_box',
        trans._('Scale Bar Box'),
        'scale_bar',
        'box',
    ),
    (
        'napari.window.view.toggle_viewer_scale_bar_colored',
        trans._('Scale Bar Colored'),
        'scale_bar',
        'colored',
    ),
    (
        'napari.window.view.toggle_viewer_scale_bar_ticks',
        trans._('Scale Bar Ticks'),
        'scale_bar',
        'ticks',
    ),
]

MENUID_DICT = {'axes': MenuId.VIEW_AXES, 'scale_bar': MenuId.VIEW_SCALEBAR}


def _tooltip_visibility_toggle() -> None:
    settings = get_settings().appearance
    settings.layer_tooltip_visibility = not settings.layer_tooltip_visibility


def _get_current_tooltip_visibility() -> bool:
    return get_settings().appearance.layer_tooltip_visibility


def _fit_to_view(viewer: ViewerModel) -> None:
    viewer.fit_to_view()


def _zoom_in(viewer: ViewerModel) -> None:
    viewer.camera.zoom *= 1.5


def _zoom_out(viewer: ViewerModel) -> None:
    viewer.camera.zoom /= 1.5


def _toggle_canvas_ndim(viewer: ViewerModel) -> None:
    """Toggle the current canvas between 3D and 2D."""
    if viewer.dims.ndisplay == 2:
        viewer.dims.ndisplay = 3
    else:  # == 3
        viewer.dims.ndisplay = 2


VIEW_ACTIONS = [
    Action(
        id='napari.viewer.fit_to_view',
        title=trans._('Fit to View'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.ZOOM,
                'order': 1,
            }
        ],
        callback=_fit_to_view,
        keybindings=[StandardKeyBinding.OriginalSize],
    ),
    Action(
        id='napari.viewer.camera.zoom_in',
        title=trans._('Zoom In'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.ZOOM,
                'order': 1,
            }
        ],
        callback=_zoom_in,
        keybindings=[StandardKeyBinding.ZoomIn],
    ),
    Action(
        id='napari.viewer.camera.zoom_out',
        title=trans._('Zoom Out'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.ZOOM,
                'order': 1,
            }
        ],
        callback=_zoom_out,
        keybindings=[StandardKeyBinding.ZoomOut],
    ),
    # TODO: this could be made into a toggle setting Action subclass
    # using a similar pattern to the above ViewerToggleAction classes
    Action(
        id='napari.window.view.toggle_ndisplay',
        title=trans._('Toggle 2D/3D Camera'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.ZOOM,
                'order': 2,
            }
        ],
        callback=_toggle_canvas_ndim,
    ),
    Action(
        id='napari.window.view.toggle_layer_tooltips',
        title=trans._('Toggle Layer Tooltips'),
        menus=[
            {
                'id': MenuId.MENUBAR_VIEW,
                'group': MenuGroup.RENDER,
                'order': 10,
            }
        ],
        callback=_tooltip_visibility_toggle,
        toggled=ToggleRule(get_current=_get_current_tooltip_visibility),
    ),
]

for cmd, cmd_title, viewer_attr, sub_attr in toggle_action_details:
    VIEW_ACTIONS.append(
        ViewerModelToggleAction(
            id=cmd,
            title=cmd_title,
            viewer_attribute=viewer_attr,
            sub_attribute=sub_attr,
            menus=[{'id': MENUID_DICT[viewer_attr]}],
        )
    )



import pooch
from tifffile import imread

import napari

"""
This data comes from the MitoNet Benchmarks.

Six benchmark volumes of instance segmentation of mitochondria from diverse volume EM datasets
Narayan K , Conrad RW
DOI: https://dx.doi.org/10.6019/EMPIAR-10982

Data is stored at EMPIAR and can be explored here: https://www.ebi.ac.uk/empiar/EMPIAR-10982/

With respect to the napari async slicing work, this dataset is small enough that it performs well in synchronous mode.
"""

salivary_gland_em_path = pooch.retrieve(
    url='https://ftp.ebi.ac.uk/empiar/world_availability/10982/data/mito_benchmarks/salivary_gland/salivary_gland_em.tif',
    known_hash='222f50dd8fd801a84f118ce71bc735f5c54f1a3ca4d98315b27721ae499bff94',
    progressbar=True
)

salivary_gland_mito_path = pooch.retrieve(
    url='https://ftp.ebi.ac.uk/empiar/world_availability/10982/data/mito_benchmarks/salivary_gland/salivary_gland_mito.tif',
    known_hash='95247d952a1dd0f7b37da1be95980b598b590e4777065c7cd877ab67cb63c5eb',
    progressbar=True
)

salivary_gland_em = imread(salivary_gland_em_path)
salivary_gland_mito = imread(salivary_gland_mito_path)

viewer = napari.Viewer()
layer = viewer.add_image(salivary_gland_em)
viewer.add_labels(salivary_gland_mito)

napari.run()



from functools import partial

import numpy as np

import napari
from napari.layers.shapes._shapes_models import (
    Ellipse,
    Line,
    Path,
    Polygon,
    Rectangle,
)


def path_length(vertices, connect_last=False):
    segment_lengths = np.linalg.norm(
        vertices - np.roll(vertices, -1, axis=0), axis=1
    )
    if connect_last:
        return np.sum(segment_lengths)
    return np.sum(segment_lengths[:-1])


def polygon_area(vertices):
    coords = np.asarray(vertices)
    n, _dims = coords.shape

    if n < 3:
        return 0

    area = 0.0

    # use a reference vertex to make the polygon out of a "triangle fan"
    ref = coords[0]
    for i in range(1, n - 1):
        u = coords[i] - ref
        v = coords[i + 1] - ref
        uu = np.dot(u, u)
        vv = np.dot(v, v)
        uv = np.dot(u, v)
        cross_sq = uu * vv - uv**2
        triangle_area = 0.5 * np.sqrt(max(cross_sq, 0))
        area += triangle_area

    return area


def rectangle_area(vertices):
    return np.linalg.norm(vertices[1] - vertices[0]) * np.linalg.norm(
        vertices[2] - vertices[1]
    )


def ellipse_area(vertices):
    return rectangle_area(vertices) * np.pi / 4


def ellipse_perimeter(vertices):
    # no closed form exists, so we use Ramanujan's 2nd approximation
    # https://en.wikipedia.org/wiki/Perimeter_of_an_ellipse#Ramanujan's_approximations
    a = np.linalg.norm(vertices[1] - vertices[0]) / 2
    b = np.linalg.norm(vertices[2] - vertices[1]) / 2
    h = (a - b) ** 2 / (a + b) ** 2
    return np.pi * (a + b) * (1 + (3 * h) / (10 + np.sqrt(4 - 3 * h)))


def update_features_with_measures(shapes_layer, event=None):
    for i, s in enumerate(shapes_layer._data_view.shapes):
        if isinstance(s, Polygon):
            a = polygon_area(s.data)
            p = path_length(s.data, connect_last=True)
        elif isinstance(s, Rectangle):
            a = rectangle_area(s.data)
            p = path_length(s.data, connect_last=True)
        elif isinstance(s, Ellipse):
            a = ellipse_area(s.data)
            p = ellipse_perimeter(s.data)
        elif isinstance(s, Path | Line):
            a = 0
            p = path_length(s.data, connect_last=False)
        shapes_layer.features.loc[i, ['_perimeter', '_area']] = p, a

    shapes_layer.refresh_text()


def get_connected_callback(shapes_layer):
    """Get the connected measure callback if present."""
    for callback in shapes_layer.events.set_data.callbacks:
        f = callback[0]() if isinstance(callback, tuple) else callback
        if isinstance(f, partial) and f.func is update_features_with_measures:
            return f
    return None


def toggle_shape_measures(shapes_layer: napari.layers.Shapes) -> None:
    """Toggle between updating and displaying measures and not."""
    if not get_connected_callback(shapes_layer):
        shapes_layer.features[['_perimeter', '_area']] = 0.0
        shapes_layer.feature_defaults = (
            shapes_layer.feature_defaults.to_dict().update(
                {'_perimeter': 0.0, '_area': 0.0}
            )
        )

        shapes_layer.events.set_data.connect(
            partial(update_features_with_measures, shapes_layer)
        )
        update_features_with_measures(shapes_layer)

        shapes_layer.text = 'P = {_perimeter:.3g}\nA = {_area:.3g}'
    else:
        shapes_layer.events.set_data.disconnect(
            get_connected_callback(shapes_layer)
        )

        shapes_layer.text = None

        shapes_layer.features = shapes_layer.features.drop(
            columns=['_perimeter', '_area']
        )
        # feature_defaults are dropped automatically



"""
Add labels
==========

Display a labels layer above of an image layer using the ``add_labels`` and
``add_image`` APIs

.. tags:: layers, visualization-basic
"""
from skimage import data
from skimage.filters import threshold_otsu
from skimage.measure import label
from skimage.morphology import closing, remove_small_objects, square
from skimage.segmentation import clear_border

import napari

image = data.coins()[50:-50, 50:-50]

# apply threshold
thresh = threshold_otsu(image)
bw = closing(image > thresh, square(4))

# remove artifacts connected to image border
cleared = remove_small_objects(clear_border(bw), 20)

# label image regions
label_image = label(cleared).astype('uint8')

# initialise viewer with coins image
viewer = napari.Viewer()
layer = viewer.add_image(image, name='coins', rgb=False)

# add the labels
label_layer = viewer.add_labels(label_image, name='segmentation')

if __name__ == '__main__':
    napari.run()



"""
Method to get napari style in magicgui based windows
====================================================

Example how to embed magicgui widget in dialog to inherit style
from main napari window.

.. tags::  gui, interactivity
"""

from magicgui import magicgui
from qtpy.QtWidgets import (
    QDialog,
    QGridLayout,
    QLabel,
    QPushButton,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

import napari
from napari.qt import get_stylesheet
from napari.settings import get_settings


# The magicgui widget shown by selecting the 'Show widget' button of MyWidget
@magicgui
def sample_add(a: int, b: int) -> int:
    return a + b

def change_style():
    sample_add.native.setStyleSheet(get_stylesheet(get_settings().appearance.theme))


get_settings().appearance.events.theme.connect(change_style)
change_style()


class MyDialog(QDialog):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.first_input = QSpinBox()
        self.second_input = QSpinBox()
        self.btn = QPushButton('Add')
        layout = QGridLayout()
        layout.addWidget(QLabel('first input'), 0, 0)
        layout.addWidget(self.first_input, 0, 1)
        layout.addWidget(QLabel('second input'), 1, 0)
        layout.addWidget(self.second_input, 1, 1)
        layout.addWidget(self.btn, 2, 0, 1, 2)
        self.setLayout(layout)
        self.btn.clicked.connect(self.run)

    def run(self):
        print('run', self.first_input.value() + self.second_input.value())
        self.close()

class MyWidget(QWidget):
    def __init__(self) -> None:
        super().__init__()
        self.btn1 = QPushButton('Show dialog')
        self.btn1.clicked.connect(self.show_dialog)
        self.btn2 = QPushButton('Show widget')
        self.btn2.clicked.connect(self.show_widget)
        self.layout = QVBoxLayout()
        self.layout.addWidget(self.btn1)
        self.layout.addWidget(self.btn2)
        self.setLayout(self.layout)

    def show_dialog(self):
        dialog = MyDialog(self)
        dialog.exec_()

    def show_widget(self):
        sample_add.show()



viewer = napari.Viewer()

widget = MyWidget()
viewer.window.add_dock_widget(widget, area='right')
napari.run()



"""
Image points 3D
===============

Display points overlaid on a 3D image

.. tags:: visualization-nD
"""
from skimage import data, feature, filters

import napari

cells = data.cells3d()
nuclei = cells[:, 1]
smooth = filters.gaussian(nuclei, sigma=10)
pts = feature.peak_local_max(smooth)
viewer = napari.Viewer(ndisplay=3)
membranes, nuclei = viewer.add_image(
    cells, channel_axis=1, name=['membranes', 'nuclei']
)
viewer.add_points(pts)
viewer.camera.angles = (10, -20, 130)

if __name__ == '__main__':
    napari.run()



"""
Action manager
==============

.. tags:: gui, experimental
"""
from random import shuffle

import numpy as np
from skimage import data

import napari
from napari._qt.widgets.qt_viewer_buttons import QtViewerPushButton
from napari.components import ViewerModel
from napari.utils.action_manager import action_manager


def rotate45(viewer: napari.Viewer):
    """
    Rotate layer 0 of the viewer by 45º

    Parameters
    ----------
    viewer : napari.Viewer
        active (unique) instance of the napari viewer

    Notes
    -----
    The `viewer` parameter needs to be named `viewer`, the action manager will
    infer that we need an instance of viewer.
    """
    angle = np.pi / 4
    from numpy import cos, sin

    r = np.array([[cos(angle), -sin(angle)], [sin(angle), cos(angle)]])
    layer = viewer.layers[0]
    layer.rotate = layer.rotate @ r


# create the viewer with an image
viewer = napari.Viewer()
layer = viewer.add_image(data.astronaut(), rgb=True)

layer_buttons = viewer.window.qt_viewer.layerButtons

# Button do not need to do anything, just need to be pretty; all the action
# binding and (un) binding will be done with the action manager, idem for
# setting the tooltip.
rot_button = QtViewerPushButton('warning')
layer_buttons.layout().insertWidget(3, rot_button)


def register_action():
    # Here we pass ViewerModel as the KeymapProvider as we want it to handle the shortcuts.
    # we could also pass none and bind the shortcuts at the window level - though we
    # are trying to not change the KeymapProvider API too much for now.
    # we give an action name to the action for configuration purposes as we need
    # it to be storable in json.

    # By convention (may be enforced later), we do give an action name which is prefixed
    # by the name of the package it is defined in, here napari,
    action_manager.register_action(
        name='napari:rotate45',
        command=rotate45,
        description='Rotate layer 0 by 45deg',
        keymapprovider=ViewerModel,
    )


def bind_shortcut():
    # note that the tooltip of the corresponding button will be updated to
    # remove the shortcut.
    action_manager.unbind_shortcut('napari:reset_view')  # Control-R
    action_manager.bind_shortcut('napari:rotate45', 'Control-R')


def bind_button():
    action_manager.bind_button('napari:rotate45', rot_button)


# we can all bind_shortcut or register_action or bind_button in any order;
# this let us configure shortcuts even if plugins are loaded / unloaded.
callbacks = [register_action, bind_shortcut, bind_button]

shuffle(callbacks)
for c in callbacks:
    print('calling', c)
    c()


# We can set the action manager in debug mode, to help us figure out which
# button is triggering which action. This will update the tooltips of the buttons
# to include the name of the action in between square brackets.

action_manager._debug(True)

# Let's also modify some existing shortcuts, by unbinding a few existing actions,
# and rebinding them with new shortcuts; below we change the add and select mode
# to be the = (same as + key on US Keyboards but without modifiers) and - keys.
# unbinding returns the old key if it exists; but we don't use it.

# in practice you likely don't need to modify the shortcuts this way as it will
# be implemented in settings, though you could imagine a plugin that would
# allow toggling between many keymaps.

settings = {
        'napari:activate_points_add_mode' : '=',
        'napari:activate_points_select_mode': '-',
}


for action, key in settings.items():
   _old_shortcut = action_manager.unbind_shortcut(action)
   action_manager.bind_shortcut(action, key)

if __name__ == '__main__':
    napari.run()



"""
Points over time
================

.. tags:: visualization-advanced
"""
import dask.array as da
import numpy as np

import napari

image4d = da.random.random(
        (4000, 32, 256, 256),
        chunks=(1, 32, 256, 256),
        )
pts_coordinates = np.random.random((50000, 3)) * image4d.shape[1:]
pts_values = da.random.random((50000, 4000), chunks=(50000, 1))

viewer = napari.Viewer(ndisplay=3)
image_layer = viewer.add_image(
        image4d, opacity=0.5
        )
pts_layer = viewer.add_points(
        pts_coordinates,
        features={'value': np.asarray(pts_values[:, 0])},
        face_color='value',
        size=2,
        )


def set_pts_features(pts_layer, values_table, step):
    # step is a 4D coordinate with the current slider position for each dim
    column = step[0]  # grab the leading ("time") coordinate
    pts_layer.features['value'] = np.asarray(values_table[:, column])
    pts_layer.face_color = 'value'  # force features refresh


viewer.dims.events.current_step.connect(
        lambda event: set_pts_features(pts_layer, pts_values, event.value)
        )


if __name__ == '__main__':
    napari.run()



from typing import Optional

import numpy as np
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QHBoxLayout, QPushButton, QWidget
from superqt import QDoubleRangeSlider

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari._qt.widgets.qt_range_slider_popup import QRangeSliderPopup
from napari.layers import Image, Surface
from napari.utils._dtype import normalize_dtype
from napari.utils.events.event_utils import connect_no_arg, connect_setattr
from napari.utils.translations import trans


def range_to_decimals(range_, dtype):
    """Convert a range to decimals of precision.

    Parameters
    ----------
    range_ : tuple
        Slider range, min and then max values.
    dtype : np.dtype
        Data type of the layer. Integers layers are given integer.
        step sizes.

    Returns
    -------
    int
        Decimals of precision.
    """
    dtype = normalize_dtype(dtype)

    if np.issubdtype(dtype, np.integer):
        return 0

    # scale precision with the log of the data range order of magnitude
    # eg.   0 - 1   (0 order of mag)  -> 3 decimal places
    #       0 - 10  (1 order of mag)  -> 2 decimals
    #       0 - 100 (2 orders of mag) -> 1 decimal
    #       ≥ 3 orders of mag -> no decimals
    # no more than 64 decimals
    d_range = np.subtract(*range_[::-1])
    return min(64, max(int(3 - np.log10(d_range)), 0))


class _QDoubleRangeSlider(QDoubleRangeSlider):
    show_clim_popup = Signal()

    def mousePressEvent(self, event):
        """Update the slider, or, on right-click, pop-up an expanded slider.

        The expanded slider provides finer control, directly editable values,
        and the ability to change the available range of the sliders.

        Parameters
        ----------
        event : napari.utils.event.Event
            The napari event that triggered this method.
        """
        if event.button() == Qt.MouseButton.RightButton:
            self.show_clim_popup.emit()
        else:
            super().mousePressEvent(event)


class QContrastLimitsPopup(QRangeSliderPopup):
    def __init__(
        self, layer: Image | Surface, parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent)

        decimals = range_to_decimals(layer.contrast_limits_range, layer.dtype)
        self.slider.setRange(*layer.contrast_limits_range)
        self.slider.setDecimals(decimals)
        self.slider.setSingleStep(10**-decimals)
        self.slider.setValue(layer.contrast_limits)

        connect_setattr(self.slider.valueChanged, layer, 'contrast_limits')
        connect_setattr(
            self.slider.rangeChanged, layer, 'contrast_limits_range'
        )

        def reset():
            layer.reset_contrast_limits()
            layer.contrast_limits_range = layer.contrast_limits
            decimals_ = range_to_decimals(
                layer.contrast_limits_range, layer.dtype
            )
            self.slider.setDecimals(decimals_)
            self.slider.setSingleStep(10**-decimals_)

        reset_btn = QPushButton('reset')
        reset_btn.setObjectName('reset_clims_button')
        reset_btn.setToolTip(trans._('Autoscale contrast to data range'))
        reset_btn.setFixedWidth(45)
        reset_btn.clicked.connect(reset)
        self._layout.addWidget(
            reset_btn, alignment=Qt.AlignmentFlag.AlignBottom
        )

        # the "full range" button doesn't do anything if it's not an
        # unsigned integer type (it's unclear what range should be set)
        # so we don't show create it at all.
        if np.issubdtype(normalize_dtype(layer.dtype), np.integer):
            range_btn = QPushButton('full range')
            range_btn.setObjectName('full_clim_range_button')
            range_btn.setToolTip(
                trans._('Set contrast range to full bit-depth')
            )
            range_btn.setFixedWidth(75)
            range_btn.clicked.connect(layer.reset_contrast_limits_range)
            self._layout.addWidget(
                range_btn, alignment=Qt.AlignmentFlag.AlignBottom
            )


class AutoScaleButtons(QWidget):
    def __init__(
        self, layer: Image | Surface, parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)

        self.setLayout(QHBoxLayout())
        self.layout().setSpacing(2)
        self.layout().setContentsMargins(0, 0, 0, 0)
        once_btn = QPushButton(trans._('once'))
        once_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        auto_btn = QPushButton(trans._('continuous'))
        auto_btn.setCheckable(True)
        auto_btn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        once_btn.clicked.connect(lambda: auto_btn.setChecked(False))
        connect_no_arg(once_btn.clicked, layer, 'reset_contrast_limits')
        connect_setattr(auto_btn.toggled, layer, '_keep_auto_contrast')
        connect_no_arg(auto_btn.clicked, layer, 'reset_contrast_limits')

        self.layout().addWidget(once_btn)
        self.layout().addWidget(auto_btn)

        # just for testing
        self._once_btn = once_btn
        self._auto_btn = auto_btn


class QtContrastLimitsControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer contrast
    limits/autocontrast and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Image | napari.layers.Surface
        An instance of a napari layer.

    Attributes
    ----------
    auto_scale_bar : AutoScaleButtons
        Widget to wrap push buttons related with the layer auto-contrast funtionality.
    auto_scale_bar_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the auto-contrast functionality widget.
    clim_popup : napari._qt.qt_range_slider_popup.QRangeSliderPopup
        Popup widget launching the contrast range slider.
    contrast_limits_slider : _QDoubleRangeSlider
        Slider controlling current constrast limits of the layer.
    contrast_limits_slider_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the constrast limits chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Image | Surface) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.contrast_limits.connect(
            self._on_contrast_limits_change
        )
        self._layer.events.contrast_limits_range.connect(
            self._on_contrast_limits_range_change
        )

        # Setup widgets
        self.auto_scale_bar = AutoScaleButtons(layer, parent)
        self.auto_scale_bar_label = QtWrappedLabel(trans._('auto-contrast:'))
        self.contrast_limits_slider = _QDoubleRangeSlider(
            Qt.Orientation.Horizontal,
        )
        self.contrast_limits_slider.show_clim_popup.connect(
            self.show_clim_popup
        )
        decimals = range_to_decimals(
            self._layer.contrast_limits_range, self._layer.dtype
        )
        self.contrast_limits_slider.setRange(
            *self._layer.contrast_limits_range
        )
        self.contrast_limits_slider.setSingleStep(10**-decimals)
        self.contrast_limits_slider.setValue(self._layer.contrast_limits)
        self.contrast_limits_slider.setToolTip(
            trans._('Right click for detailed slider popup.')
        )

        self.clim_popup = None

        connect_setattr(
            self.contrast_limits_slider.valueChanged,
            self._layer,
            'contrast_limits',
        )
        connect_setattr(
            self.contrast_limits_slider.rangeChanged,
            self._layer,
            'contrast_limits_range',
        )

        self.contrast_limits_slider_label = QtWrappedLabel(
            trans._('contrast limits:')
        )

    def show_clim_popup(self):
        self.clim_popup = QContrastLimitsPopup(self._layer, self.parent())
        self.clim_popup.setParent(self.parent())
        self.clim_popup.move_to('top', min_length=650)
        self.clim_popup.show()

    def _on_contrast_limits_change(self):
        """Receive layer model contrast limits change event and update slider."""
        with qt_signals_blocked(self.contrast_limits_slider):
            self.contrast_limits_slider.setValue(self._layer.contrast_limits)

        if self.clim_popup:
            with qt_signals_blocked(self.clim_popup.slider):
                self.clim_popup.slider.setValue(self._layer.contrast_limits)

    def _on_contrast_limits_range_change(self):
        """Receive layer model contrast limits change event and update slider."""
        with qt_signals_blocked(self.contrast_limits_slider):
            decimals = range_to_decimals(
                self._layer.contrast_limits_range, self._layer.dtype
            )
            self.contrast_limits_slider.setRange(
                *self._layer.contrast_limits_range
            )
            self.contrast_limits_slider.setSingleStep(10**-decimals)

        if self.clim_popup:
            with qt_signals_blocked(self.clim_popup.slider):
                self.clim_popup.slider.setRange(
                    *self._layer.contrast_limits_range
                )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [
            (self.contrast_limits_slider_label, self.contrast_limits_slider),
            (self.auto_scale_bar_label, self.auto_scale_bar),
        ]
